# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/tautology.py

"""The Tautology Theorem and its implications."""

from typing import List, Sequence, Union
from functools import reduce
from mltp.logic_utils import frozendict
from .syntax import *
from .semantics import *
from .proofs import *
from .axiomatic_systems import *
from .deduction import *
from .operators import *

def formulas_capturing_model(model: Model) -> List[Formula]:
    """Computes the formulas that capture the given model: ``'``\ `x`\ ``'``
    for each variable name `x` that is assigned the value ``True`` in the given
    model, and ``'~``\ `x`\ ``'`` for each variable name `x` that is assigned
    the value ``False``.

    Parameters:
        model: model to construct the formulas for.

    Returns:
        A list of the constructed formulas, ordered alphabetically by variable
        name.

    Examples:
        >>> formulas_capturing_model({'p2': False, 'p1': True, 'q': True})
        [p1, ~p2, q]
    """
    assert is_model(model)
    list1 = []
    variables = sorted(model.keys())
    for i in variables:
        if model[i]:
            list1.append(Formula.parse(i))
        else:
            list1.append(Formula('~',Formula.parse(i)))
    return list1
    # Task 6.1a

def prove_in_model(formula: Formula, model:Model) -> Proof:
    """Either proves the given formula or proves its negation, from the formulas
    that capture the given model.

    Parameters:
        formula: formula that contains no constants or operators beyond ``'->'``
            and ``'~'``, whose affirmation or negation is to prove.
        model: model from whose formulas to prove.

    Returns:
        If `formula` evaluates to ``True`` in the given model, then a valid
        proof of `formula`; otherwise a valid proof of ``'~``\ `formula`\ ``'``.
        The returned proof is from the formulas that capture the given model, in
        the order returned by `formulas_capturing_model`\ ``(``\ `model`\ ``)``,
        via `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.

    Examples:
        >>> proof = prove_in_model(Formula.parse('(p->q7)'),
        ...                        {'q7': False, 'p': False})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (p->q7)
        >>> proof.statement.assumptions
        (~p, ~q7)
        >>> proof.rules == AXIOMATIC_SYSTEM
        True

        >>> proof = prove_in_model(Formula.parse('(p->q7)'),
        ...                        {'q7': False, 'p': True})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        ~(p->q7)
        >>> proof.statement.assumptions
        (p, ~q7)
        >>> proof.rules == AXIOMATIC_SYSTEM
        True
    """
    assert formula.operators().issubset({'->', '~'})
    assert is_model(model)

    assumptions = formulas_capturing_model(model)
    rules = AXIOMATIC_SYSTEM
    root = formula.root

    if is_variable(root):
        if evaluate(formula, model):
            conclusion = formula
        else:
            conclusion = Formula('~', formula)

        line = [Proof.Line(conclusion)]
        statement = InferenceRule(assumptions, conclusion)
        proof = Proof(statement, rules, line)
        return proof

    elif root == '~':
        if evaluate(formula, model):
            return prove_in_model(formula.first, model)
        else:
            proof = prove_in_model(formula.first, model)
            consequent = Formula('~', formula)
            conditional = NN
            return prove_corollary(proof, consequent, conditional)

    elif root == '->':
        if evaluate(formula, model):
            if not evaluate(formula.first, model):
                proof = prove_in_model(Formula('~', formula.first), model)
                consequent = I2.specialize({'p': formula.first, 'q': formula.second}).conclusion.second
                conditional = I2
                return prove_corollary(proof, consequent, conditional)

            else:
                proof = prove_in_model(formula.second, model)
                consequent = I1.specialize({'p': formula.first, 'q': formula.second}).conclusion.second
                conditional = I1
                return prove_corollary(proof, consequent, conditional)

        else:
            proof1 = prove_in_model(formula.first, model)
            proof2 = prove_in_model(Formula('~', formula.second), model)
            consequent = Formula('~', formula)
            return combine_proofs(proof1, proof2, consequent, NI)
    # Task 6.1b

def reduce_assumption(proof_from_affirmation: Proof, proof_from_negation: Proof) -> Proof:
    """Combines the two given proofs, both of the same formula `conclusion` and
    from the same assumptions except that the last assumption of the latter is
    the negation of that of the former, into a single proof of `conclusion` from
    only the common assumptions.

    Parameters:
        proof_from_affirmation: valid proof of `conclusion` from one or more
            assumptions, the last of which is an assumption `assumption`.
        proof_from_negation: valid proof of `conclusion` from the same
            assumptions and inference rules of `proof_from_affirmation`, but
            with the last assumption being ``'~``\ `assumption`\ ``'`` instead
            of `assumption`.

    Returns:
        A valid proof of `conclusion` from only the assumptions common to the
        given proofs (i.e., without the last assumption of each), via the same
        inference rules of the given proofs and in addition
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.R`.

    Examples:
        If `proof_from_affirmation` is of
                       ``['p', '~q', 'r'] ==> '(p&(r|~r))'``, then `proof_from_negation` must be of
                       ``['p', '~q', '~r'] ==> '(p&(r|~r))'`` and the returned proof is of
                       ``['p', '~q'] ==> '(p&(r|~r))'``.
    """
    assert proof_from_affirmation.is_valid()
    assert proof_from_negation.is_valid()
    assert proof_from_affirmation.statement.conclusion == proof_from_negation.statement.conclusion
    assert len(proof_from_affirmation.statement.assumptions) > 0
    assert len(proof_from_negation.statement.assumptions) > 0
    assert proof_from_affirmation.statement.assumptions[:-1] == proof_from_negation.statement.assumptions[:-1]
    assert Formula('~', proof_from_affirmation.statement.assumptions[-1]) == proof_from_negation.statement.assumptions[-1]
    assert proof_from_affirmation.rules == proof_from_negation.rules

    return combine_proofs(remove_assumption(proof_from_affirmation),
                          remove_assumption(proof_from_negation),
                          proof_from_affirmation.statement.conclusion,
                          R)

    # Task 6.2

def prove_tautology(tautology: Formula, model: Model = frozendict()) -> Proof:
    """Proves the given tautology from the formulas that capture the given
    model.

    Parameters:
        tautology: tautology that contains no constants or operators beyond
            ``'->'`` and ``'~'``, to prove.
        model: model over a (possibly empty) prefix (with respect to the
            alphabetical order) of the variable names of `tautology`, from whose
            formulas to prove.

    Returns:
        A valid proof of the given tautology from the formulas that capture the
        given model, in the order returned by
        `formulas_capturing_model`\ ``(``\ `model`\ ``)``, via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.

    Examples:
        >>> proof = prove_tautology(Formula.parse('(~(p->p)->q)'),
        ...                         {'p': True, 'q': False})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (~(p->p)->q)
        >>> proof.statement.assumptions
        (p, ~q)
        >>> proof.rules == AXIOMATIC_SYSTEM
        True

        >>> proof = prove_tautology(Formula.parse('(~(p->p)->q)'))
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (~(p->p)->q)
        >>> proof.statement.assumptions
        ()
        >>> proof.rules == AXIOMATIC_SYSTEM
        True
    """
    assert is_tautology(tautology)
    assert tautology.operators().issubset({'->', '~'})
    assert is_model(model)
    assert sorted(tautology.variables())[:len(model)] == sorted(model.keys())

    if sorted(tautology.variables()) == sorted(model.keys()):
        return prove_in_model(tautology, model)
    else:
        model = dict(model)
        model_copy = {i: model[i] for i in model.keys()}
        next_variable = sorted(tautology.variables())[len(model)]

        model.update({next_variable: True})
        model_copy.update({next_variable: False})

        return reduce_assumption(prove_tautology(tautology, model),
                                 prove_tautology(tautology, model_copy))
    # Task 6.3a

def proof_or_counterexample(formula: Formula) -> Union[Proof, Model]:
    """Either proves the given formula or finds a model in which it does not
    hold.

    Parameters:
        formula: formula that contains no constants or operators beyond ``'->'``
            and ``'~'``, to either prove or find a counterexample for.

    Returns:
        If the given formula is a tautology, then an assumptionless proof of the
        formula via `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`,
        otherwise a model in which the given formula does not hold.
    """
    assert formula.operators().issubset({'->', '~'})
    if is_tautology(formula):
        return prove_tautology(formula)
    else:
        for i in all_models(formula.variables()):
            if not evaluate(formula, i):
                return i
    # Task 6.3b

def encode_as_formula(rule: InferenceRule) -> Formula:
    """Encodes the given inference rule as a formula consisting of a chain of
    implications.

    Parameters:
        rule: inference rule to encode.

    Returns:
        The formula encoding the given rule.

    Examples:
        >>> encode_as_formula(InferenceRule([Formula('p1'), Formula('p2'),
        ...                                  Formula('p3'), Formula('p4')],
        ...                                 Formula('q')))
        (p1->(p2->(p3->(p4->q))))

        >>> encode_as_formula(InferenceRule([], Formula('q')))
        q
    """

    if rule.assumptions:
        new_assumptions = rule.assumptions[:-1]
        new_conclusion = Formula('->', rule.assumptions[-1], rule.conclusion)
        return encode_as_formula(InferenceRule(new_assumptions, new_conclusion))
    else:
        return rule.conclusion
    # Task 6.4a

def prove_sound_inference(rule: InferenceRule) -> Proof:
    """Proves the given sound inference rule.

    Parameters:
        rule: sound inference rule whose assumptions and conclusion contain no
            constants or operators beyond ``'->'`` and ``'~'``, to prove.

    Returns:
        A valid proof of the given sound inference rule via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.
    """
    assert is_sound_inference(rule)
    for formula in {rule.conclusion}.union(rule.assumptions):
        assert formula.operators().issubset({'->', '~'})

    encoding = encode_as_formula(rule)

    rules = AXIOMATIC_SYSTEM
    lines = list(prove_tautology(encoding).lines)

    n = len(lines)
    for assumption in rule.assumptions:
        encoding = encoding.second
        lines.append(Proof.Line(assumption))
        lines.append(Proof.Line(encoding, MP, (n, n-1)))
        n += 2

    return Proof(rule, rules, lines)
    # Task 6.4b

def model_or_inconsistency(formulae: Sequence[Formula]) -> Union[Model, Proof]:
    """Either finds a model in which all the given formulas hold, or proves
    ``'~(p->p)'`` from these formulas.

    Parameters:
        formulas: formulas that use only the operators ``'->'`` and ``'~'``, to
            either find a model of, or prove ``'~(p->p)'`` from.

    Returns:
        A model in which all of the given formulas hold if such exists,
        otherwise a valid proof of ``'~(p->p)'`` from the given formulas via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.
    """
    for formula in formulae:
        assert formula.operators().issubset({'->', '~'})

    variables = reduce(lambda x, y: x | y, (set(formula.variables()) for formula in formulae))
    models = all_models(variables)

    for formula in formulae:
        models = [model for model in models if evaluate(formula, model)]

    if models:
        return models[0]
    else:
        return prove_sound_inference(InferenceRule(formulae, Formula.parse('~(p->p)')))
    # Task 6.5

def prove_in_model_full(formula: Formula, model: Model) -> Proof:
    """Either proves the given formula or proves its negation, from the formulas
    that capture the given model.

    Parameters:
        formula: formula that contains no operators beyond ``'->'``, ``'~'``,
            ``'&'``, and ``'|'`` (and may contain constants), whose affirmation
            or negation is to prove.
        model: model from whose formulas to prove.

    Returns:
        If `formula` evaluates to ``True`` in the given model, then a valid
        proof of `formula`; otherwise a valid proof of ``'~``\ `formula`\ ``'``.
        The returned proof is from the formulas that capture the given model, in
        the order returned by `formulas_capturing_model`\ ``(``\ `model`\ ``)``,
        via `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM_FULL`.

    Examples:
        >>> proof = prove_in_model_full(Formula.parse('(p&q7)'),
        ...                             {'q7': True, 'p': True})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (p&q7)
        >>> proof.statement.assumptions
        (p, q7)
        >>> proof.rules == AXIOMATIC_SYSTEM_FULL
        True

        >>> proof = prove_in_model_full(Formula.parse('(p&q7)'),
        ...                             {'q7': False, 'p': True})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        ~(p&q7)
        >>> proof.statement.assumptions
        (p, ~q7)
        >>> proof.rules == AXIOMATIC_SYSTEM_FULL
        True
    """
    assert formula.operators().issubset({'T', 'F', '->', '~', '&', '|'})
    assert is_model(model)

    assumptions = formulas_capturing_model(model)
    rules = AXIOMATIC_SYSTEM_FULL
    root = formula.root

    if is_variable(root):
        if evaluate(formula, model):
            conclusion = formula
        else:
            conclusion = Formula('~', formula)

        line = [Proof.Line(conclusion)]
        statement = InferenceRule(assumptions, conclusion)
        proof = Proof(statement, rules, line)
        return proof

    elif root == '~':
        if evaluate(formula, model):
            return prove_in_model_full(formula.first, model)
        else:
            proof = prove_in_model_full(formula.first, model)
            consequent = Formula('~', formula)
            conditional = NN
            return prove_corollary(proof, consequent, conditional)

    elif root == '->':
        if evaluate(formula, model):
            if not evaluate(formula.first, model):
                proof = prove_in_model_full(Formula('~', formula.first), model)
                consequent = I2.specialize({'p': formula.first, 'q': formula.second}).conclusion.second
                conditional = I2
                return prove_corollary(proof, consequent, conditional)

            else:
                proof = prove_in_model_full(formula.second, model)
                consequent = I1.specialize({'p': formula.first, 'q': formula.second}).conclusion.second
                conditional = I1
                return prove_corollary(proof, consequent, conditional)

        else:
            proof1 = prove_in_model_full(formula.first, model)
            proof2 = prove_in_model_full(Formula('~', formula.second), model)
            consequent = Formula('~', formula)
            return combine_proofs(proof1, proof2, consequent, NI)

    elif root == '&':
        if evaluate(formula, model):
            proof1 = prove_in_model_full(formula.first, model)
            proof2 = prove_in_model_full(formula.second, model)
            consequent = formula
            double_conditional = A
            return combine_proofs(proof1, proof2, consequent, double_conditional)

        else:
            if evaluate(formula.first, model):
                proof = prove_in_model_full(Formula('~', formula.second), model)
                consequent = Formula('~', formula)
                conditional = NA1
                return prove_corollary(proof, consequent, conditional)

            else:
                proof = prove_in_model_full(Formula('~', formula.first), model)
                consequent = Formula('~', formula)
                conditional = NA2
                return prove_corollary(proof, consequent, conditional)

    elif root == '|':
        if evaluate(formula, model):
            if evaluate(formula.first, model):
                proof = prove_in_model_full(formula.first, model)
                consequent = formula
                conditional = O2
                return prove_corollary(proof, consequent, conditional)

            else:
                proof = prove_in_model_full(formula.second, model)
                consequent = formula
                conditional = O1
                return prove_corollary(proof, consequent, conditional)
        else:
            proof1 = prove_in_model_full(Formula('~', formula.first), model)
            proof2 = prove_in_model_full(Formula('~', formula.second), model)
            consequent = Formula('~', formula)
            double_conditional = NO
            return combine_proofs(proof1, proof2, consequent, double_conditional)

    elif root == 'T':
        line = [Proof.Line(formula, T, ())]
        statement = T
        return Proof(statement, rules, line)

    else:
        assert root == 'F'
        line = [Proof.Line(Formula('~', formula), NF, ())]
        statement = NF
        return Proof(statement, rules, line)
    # Optional Task 6.6

def prove_tautology_full(tautology: Formula, model: Model = frozendict()) -> Proof:
    """Proves the given tautology from the formulas that capture the given
    model.

    Parameters:
        tautology: tautology that contains no constants or operators beyond
            ``'->'`` and ``'~'``, to prove.
        model: model over a (possibly empty) prefix (with respect to the
            alphabetical order) of the variable names of `tautology`, from whose
            formulas to prove.

    Returns:
        A valid proof of the given tautology from the formulas that capture the
        given model, in the order returned by
        `formulas_capturing_model`\ ``(``\ `model`\ ``)``, via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.

    Examples:
        >>> proof = prove_tautology(Formula.parse('(~(p->p)->q)'),
        ...                         {'p': True, 'q': False})
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (~(p->p)->q)
        >>> proof.statement.assumptions
        (p, ~q)
        >>> proof.rules == AXIOMATIC_SYSTEM
        True

        >>> proof = prove_tautology(Formula.parse('(~(p->p)->q)'))
        >>> proof.is_valid()
        True
        >>> proof.statement.conclusion
        (~(p->p)->q)
        >>> proof.statement.assumptions
        ()
        >>> proof.rules == AXIOMATIC_SYSTEM
        True
    """
    assert is_tautology(tautology)
    assert tautology.operators().issubset({'T', 'F', '->', '~', '&', '|'})
    assert is_model(model)
    assert sorted(tautology.variables())[:len(model)] == sorted(model.keys())

    if sorted(tautology.variables()) == sorted(model.keys()):
        return prove_in_model_full(tautology, model)
    else:
        model = dict(model)
        model_copy = {i: model[i] for i in model.keys()}
        next_variable = sorted(tautology.variables())[len(model)]

        model.update({next_variable: True})
        model_copy.update({next_variable: False})

        return reduce_assumption(prove_tautology_full(tautology, model),
                                 prove_tautology_full(tautology, model_copy))
    # Task 6.3a

def proof_or_counterexample_full(formula: Formula) -> Union[Proof, Model]:
    """Either proves the given formula or finds a model in which it does not
    hold.

    Parameters:
        formula: formula that contains no constants or operators beyond ``'->'``
            and ``'~'``, to either prove or find a counterexample for.

    Returns:
        If the given formula is a tautology, then an assumptionless proof of the
        formula via `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`,
        otherwise a model in which the given formula does not hold.
    """
    assert formula.operators().issubset({'T', 'F', '->', '~', '&', '|'})
    if is_tautology(formula):
        return prove_tautology_full(formula)
    else:
        for i in all_models(formula.variables()):
            if not evaluate(formula, i):
                return i
    # Task 6.3b

def prove_sound_inference_full(rule: InferenceRule) -> Proof:
    """Proves the given sound inference rule.

    Parameters:
        rule: sound inference rule whose assumptions and conclusion contain no
            constants or operators beyond ``'->'`` and ``'~'``, to prove.

    Returns:
        A valid proof of the given sound inference rule via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.
    """
    assert is_sound_inference(rule)
    for formula in {rule.conclusion}.union(rule.assumptions):
        assert formula.operators().issubset({'T', 'F', '->', '~', '&', '|'})

    encoding = encode_as_formula(rule)

    rules = AXIOMATIC_SYSTEM
    lines = list(prove_tautology_full(encoding).lines)

    n = len(lines)
    for assumption in rule.assumptions:
        encoding = encoding.second
        lines.append(Proof.Line(assumption))
        lines.append(Proof.Line(encoding, MP, (n, n-1)))
        n += 2

    return Proof(rule, rules, lines)
    # Task 6.4b

def model_or_inconsistency_full(formulae: Sequence[Formula]) -> Union[Model, Proof]:
    """Either finds a model in which all the given formulas hold, or proves
    ``'~(p->p)'`` from these formulas.

    Parameters:
        formulas: formulas that use only the operators ``'->'`` and ``'~'``, to
            either find a model of, or prove ``'~(p->p)'`` from.

    Returns:
        A model in which all of the given formulas hold if such exists,
        otherwise a valid proof of ``'~(p->p)'`` from the given formulas via
        `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.
    """
    for formula in formulae:
        assert formula.operators().issubset({'T', 'F', '->', '~', '&', '|'})

    variables = reduce(lambda x, y: x | y, (set(formula.variables()) for formula in formulae))
    models = all_models(variables)

    for formula in formulae:
        models = [model for model in models if evaluate(formula, model)]

    if models:
        return models[0]
    else:
        return prove_sound_inference_full(InferenceRule(formulae, Formula.parse('~(p->p)')))
    # Task 6.5

# variables = reduce((lambda x, y: x | y), (set(formula.variables()) for formula in formulae))
# models = all_models(variables)

# for formula in formulae:
#     models = (model for model in models if evaluate(formula, model))
# if models:
#     print(next(models))

#--------------

# formula = to_implies_not(Formula.parse('(((((((x11|x12)|x13)&(~x11|~x12))&(~x11|~x13))&(~x12|~x13))&(((((x21|x22)|x23)&(~x21|~x22))&(~x21|~x23))&(~x22|~x23)))&(((((x31|x32)|x33)&(~x31|~x32))&(~x31|~x33))&(~x32|~x33)))'))
# allmodels = all_models(formula.variables())
# for i in range(10):
#     proof1 = prove_in_model(formula, next(allmodels))
#     print(proof1.statement, '\n')

#--------------

# formula = to_implies_not(Formula.parse('((((((((((((((((((((((((((((((((((((((((((((((((x11|x12)|x13)&(~x11|~x12))&(~x11|~x13))&(~x12|~x13))&(((((x21|x22)|x23)&(~x21|~x22))&(~x21|~x23))&(~x22|~x23)))&(((((x31|x32)|x33)&(~x31|~x32))&(~x31|~x33))&(~x32|~x33)))&(((((x41|x42)|x43)&(~x41|~x42))&(~x41|~x43))&(~x42|~x43)))&(((((x51|x52)|x53)&(~x51|~x52))&(~x51|~x53))&(~x52|~x53)))&(((((x61|x62)|x63)&(~x61|~x62))&(~x61|~x63))&(~x62|~x63)))&(((((x71|x72)|x73)&(~x71|~x72))&(~x71|~x73))&(~x72|~x73)))&(((((x81|x82)|x83)&(~x81|~x82))&(~x81|~x83))&(~x82|~x83)))&(~x21|~x41))&(~x22|~x42))&(~x23|~x43))&(~x51|~x81))&(~x52|~x82))&(~x53|~x83))&(~x41|~x61))&(~x42|~x62))&(~x43|~x63))&(~x11|~x41))&(~x12|~x42))&(~x13|~x43))&(~x51|~x71))&(~x52|~x72))&(~x53|~x73))&(~x21|~x31))&(~x22|~x32))&(~x23|~x33))&(~x11|~x71))&(~x12|~x72))&(~x13|~x73))&(~x41|~x51))&(~x42|~x52))&(~x43|~x53))&(~x71|~x81))&(~x72|~x82))&(~x73|~x83))&(~x11|~x61))&(~x12|~x62))&(~x13|~x63))&(~x41|~x71))&(~x42|~x72))&(~x43|~x73))&(~x31|~x51))&(~x32|~x52))&(~x33|~x53))'))
# print(formula, '\n\n\n')
# assignment = {'x11':True, 'x12':False, 'x13':False, 'x21':True, 'x22':False, 'x23':False, 'x31':False, 'x32':True, 'x33':False, 'x41':False, 'x42':True, 'x43':False, 'x51':True, 'x52':False, 'x53':False, 'x61':False, 'x62':False, 'x63':True, 'x71':False, 'x72':False, 'x73':True, 'x81':False, 'x82':True, 'x83':False}
# proof1 = prove_in_model(formula, assignment)
# print('\n', len(proof1.lines), len((proof1.clean()).lines))



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/deduction.py

"""Useful proof manipulation maneuvers in Propositional Logic."""
from .syntax import *
from .proofs import *
from .axiomatic_systems import *

def prove_corollary(antecedent_proof: Proof, consequent: Formula,
                    conditional: InferenceRule) -> Proof:
    """Converts the given proof of a formula `antecedent` to a proof of the
    given formula `consequent` by using the given assumptionless inference rule
    of which ``'(``\ `antecedent`\ ``->``\ `consequent`\ ``)'`` is a
    specialization.

    Parameters:
        antecedent_proof: valid proof of `antecedent`.
        consequent: formula to prove.
        conditional: assumptionless inference rule of which the assumptionless
            inference rule with conclusion
            ``'(``\ `antecedent`\ ``->``\ `consequent`\ ``)'`` is a
            specialization.

    Returns:
        A valid proof of `consequent` from the same assumptions as the given
        proof, via the same inference rules as the given proof and in addition
        `~propositions.axiomatic_systems.MP` and `conditional`.
    """
    assert antecedent_proof.is_valid()
    assert InferenceRule([],
                         Formula('->', antecedent_proof.statement.conclusion,
                                 consequent)).is_specialization_of(conditional)

    new_statement = InferenceRule(list(antecedent_proof.statement.assumptions), consequent)

    new_rules = antecedent_proof.rules | set((MP, conditional))

    specialized_conditional = Formula('->', antecedent_proof.statement.conclusion, consequent)

    n = len(antecedent_proof.lines)

    new_lines = list(antecedent_proof.lines)
    new_lines.append(Proof.Line(specialized_conditional, conditional, []))
    new_lines.append(Proof.Line(consequent, MP, [n-1,n]))

    return Proof(new_statement, new_rules, new_lines)
    # Task 5.3a

def combine_proofs(antecedent1_proof: Proof, antecedent2_proof: Proof,
                   consequent: Formula, double_conditional: InferenceRule) -> \
        Proof:
    """Combines the given proofs of two formulas `antecedent1` and `antecedent2`
    into a proof of the given formula `consequent` by using the given
    assumptionless inference rule of which
    ``'(``\ `antecedent1`\ ``->(``\ `antecedent2`\ ``->``\ `consequent`\ ``))'``
    is a specialization.

    Parameters:
        antecedent1_proof: valid proof of `antecedent1`.
        antecedent2_proof: valid proof of `antecedent2` from the same
            assumptions and inference rules as `antecedent1_proof`.
        consequent: formula to prove.
        double_conditional: assumptionless inference rule of which the
            assumptionless inference rule with conclusion
            ``'(``\ `antecedent1`\ ``->(``\ `antecedent2`\ ``->``\ `consequent`\ ``))'``
            is a specialization.

    Returns:
        A valid proof of `consequent` from the same assumptions as the given
        proofs, via the same inference rules as the given proofs and in addition
        `~propositions.axiomatic_systems.MP` and `double_conditional`.
    """
    assert antecedent1_proof.is_valid()
    assert antecedent2_proof.is_valid()
    assert antecedent1_proof.statement.assumptions == antecedent2_proof.statement.assumptions
    assert antecedent1_proof.rules == antecedent2_proof.rules
    assert InferenceRule(
        [], Formula('->', antecedent1_proof.statement.conclusion,
        Formula('->', antecedent2_proof.statement.conclusion, consequent))
        ).is_specialization_of(double_conditional)

    new_statement = InferenceRule((antecedent1_proof.statement.assumptions + tuple(set(antecedent2_proof.statement.assumptions)-set(antecedent1_proof.statement.assumptions))),consequent)

    new_rules = antecedent1_proof.rules | antecedent2_proof.rules | set((MP, double_conditional))

    n = len(antecedent1_proof.lines)
    m = len(antecedent2_proof.lines)

    specialized_double_conditional = Formula('->', antecedent1_proof.statement.conclusion, Formula('->', antecedent2_proof.statement.conclusion, consequent))
    specialized_single_conditional = specialized_double_conditional.second

    new_lines1 = list(antecedent1_proof.lines)
    new_lines2 = [line if line.is_assumption() else Proof.Line(line.formula, line.rule, [assumption+n for assumption in line.assumptions]) for line in antecedent2_proof.lines]
    new_lines = new_lines1 + new_lines2

    new_lines.append(Proof.Line(specialized_double_conditional, double_conditional, []))
    new_lines.append(Proof.Line(specialized_single_conditional, MP, [n-1,n+m]))
    new_lines.append(Proof.Line(consequent, MP, [n+m-1, n+m+1]))

    return Proof(new_statement, new_rules, new_lines)
    # Task 5.3b

def remove_assumption(proof: Proof) -> Proof:
    """Converts the given proof of some `conclusion` formula, the last
    assumption of which is an assumption `assumption`, to a proof of
    ``'(``\ `assumption`\ ``->``\ `conclusion`\ ``)'`` from the same assumptions
    except `assumption`.

    Parameters:
        proof: valid proof to convert, with at least one assumption, via some
            set of inference rules all of which have no assumptions except
            perhaps `~propositions.axiomatic_systems.MP`.

    Returns:
        A valid proof of ``'(``\ `assumption`\ ``->``\ `conclusion`\ ``)'``
        from the same assumptions as the given proof except the last one, via
        the same inference rules as the given proof and in addition
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`, and
        `~propositions.axiomatic_systems.D`.
    """
    assert proof.is_valid()
    assert len(proof.statement.assumptions) > 0
    for rule in proof.rules:
        assert rule == MP or len(rule.assumptions) == 0

    phi = proof.statement.assumptions[-1]
    phi_I0 = Formula('->', phi, phi)

    new_lines = []
    line_index = []

    for i in range(len(proof.lines)):
        line = proof.lines[i]

        if line.formula == phi:
            line_index.append(i)
            new_lines.append(Proof.Line(phi_I0, I0, []))

        elif line.is_assumption() or len(line.rule.assumptions) == 0:
            [line_index.append(i) for j in range(3)]
            q = line.formula

            new_lines.append(line)
            new_lines.append(Proof.Line(I1.specialize({'p': phi, 'q': q}).conclusion, I1, []))
            new_lines.append(Proof.Line(Formula('->', phi, q), MP, [len(new_lines)-2, len(new_lines)-1]))

        else:
            [line_index.append(i) for j in range(3)]
            n = len(new_lines)
            q = proof.lines[line.assumptions[1]].formula.first

            new_lines.append(Proof.Line(D.specialize({'p': phi, 'q': q, 'r': line.formula}).conclusion, D, []))

            antecedent_1 = new_lines[-1].formula.first
            for line_number in range(len(new_lines)):
                if new_lines[line_number].formula == antecedent_1:
                    p1_1 = line_number
                    break
            new_lines.append(Proof.Line(new_lines[-1].formula.second, MP, [p1_1, n]))

            antecedent_2 = new_lines[-1].formula.first
            for line_number in range(len(new_lines)):
                if new_lines[line_number].formula == antecedent_2:
                    p1_2 = line_number
                    break
            new_lines.append(Proof.Line(new_lines[-1].formula.second, MP, [p1_2, n+1]))

    new_statement = InferenceRule((proof.statement.assumptions[:-1]), Formula('->',proof.statement.assumptions[-1],proof.statement.conclusion))
    new_rules = proof.rules | set((I0, D, I1, MP))

    return Proof(new_statement, new_rules, new_lines)
    # Task 5.4

def prove_from_opposites(proof_of_affirmation: Proof,
                         proof_of_negation: Proof, conclusion: Formula) -> \
        Proof:
    """Combines the given proofs of a formula `affirmation` and its negation
    ``'~``\ `affirmation`\ ``'`` into a proof of the given formula.

    Parameters:
        proof_of_affirmation: valid proof of `affirmation`.
        proof_of_negation: valid proof of ``'~``\ `affirmation`\ ``'`` from the
            same assumptions and inference rules of `proof_of_affirmation`.
        conclusion: formula to prove.

    Returns:
        A valid proof of `conclusion` from the same assumptions as the given
        proofs, via the same inference rules as the given proofs and in addition
        `~propositions.axiomatic_systems.MP` and
        `~propositions.axiomatic_systems.I2`.
    """
    assert proof_of_affirmation.is_valid()
    assert proof_of_negation.is_valid()
    assert proof_of_affirmation.statement.assumptions == proof_of_negation.statement.assumptions
    assert Formula('~', proof_of_affirmation.statement.conclusion) == proof_of_negation.statement.conclusion
    assert proof_of_affirmation.rules == proof_of_negation.rules

    return combine_proofs(proof_of_negation, proof_of_affirmation, conclusion, I2)
    # Task 5.6

def prove_by_way_of_contradiction(proof: Proof) -> Proof:
    """Converts the given proof of ``'~(p->p)'``, the last assumption of which
    is an assumption ``'~``\ `formula`\ ``'``, to a proof of `formula` from the
    same assumptions except ``'~``\ `formula`\ ``'``.

    Parameters:
        proof: valid proof of ``'~(p->p)'`` to convert, the last assumption of
            which is of the form ``'~``\ `formula`\ ``'``, via some set of
            inference rules all of which have no assumptions except perhaps
            `~propositions.axiomatic_systems.MP`.

    Returns:
        A valid proof of `formula` from the same assumptions as the given proof
        except the last one, via the same inference rules as the given proof and
        in addition `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    assert proof.is_valid()
    assert proof.statement.conclusion == Formula.parse('~(p->p)')
    assert len(proof.statement.assumptions) > 0
    assert proof.statement.assumptions[-1].root == '~'
    for rule in proof.rules:
        assert rule == MP or len(rule.assumptions) == 0

    r = proof.statement.assumptions[-1].first

    proof = remove_assumption(proof)
    proof = prove_corollary(proof, Formula('->',Formula.parse('(p->p)'),r), N)
    new_statement = InferenceRule(proof.statement.assumptions, r)

    n = len(proof.lines)
    new_lines = list(proof.lines)
    new_lines.append(Proof.Line(Formula.parse('(p->p)'), I0, []))
    new_lines.append(Proof.Line(r, MP, [n, n-1]))

    return Proof(new_statement, proof.rules, new_lines)
    # Task 5.7

# proof1 = remove_assumption(remove_assumption(parse_proof("""Proof of ['p', 'q'] ==> '(r->(r&(p&q)))' via inference rules:
#   [] ==> '(p->(q->(p&q)))'
#   ['p', '(p->q)'] ==> 'q'
#   [] ==> '(q->(p->q))'
#   [] ==> '((p->(q->r))->((p->q)->(p->r)))'
#   [] ==> '(p->p)'
# Lines:
#   0) p
#   1) (p->(r->p))    (Inference Rule [] ==> '(q->(p->q))')
#   2) (r->p)    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 0,1)
#   3) q
#   4) (q->(r->q))    (Inference Rule [] ==> '(q->(p->q))')
#   5) (r->q)    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 3,4)
#   6) p
#   7) (p->(r->p))    (Inference Rule [] ==> '(q->(p->q))')
#   8) (r->p)    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 6,7)
#   9) q
#  10) (q->(r->q))    (Inference Rule [] ==> '(q->(p->q))')
#  11) (r->q)    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 9,10)
#  12) (p->(q->(p&q)))    (Inference Rule [] ==> '(p->(q->(p&q)))')
#  13) ((p->(q->(p&q)))->(r->(p->(q->(p&q)))))    (Inference Rule [] ==> '(q->(p->q))')
#  14) (r->(p->(q->(p&q))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 12,13)
#  15) ((r->(p->(q->(p&q))))->((r->p)->(r->(q->(p&q)))))    (Inference Rule [] ==> '((p->(q->r))->((p->q)->(p->r)))')
#  16) ((r->p)->(r->(q->(p&q))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 14,15)
#  17) (r->(q->(p&q)))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 2,16)
#  18) ((r->(q->(p&q)))->((r->q)->(r->(p&q))))    (Inference Rule [] ==> '((p->(q->r))->((p->q)->(p->r)))')
#  19) ((r->q)->(r->(p&q)))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 17,18)
#  20) (r->(p&q))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 5,19)
#  21) (r->r)    (Inference Rule [] ==> '(p->p)')
#  22) (r->r)    (Inference Rule [] ==> '(p->p)')
#  23) ((r->(q->(p&q)))->((r->q)->(r->(p&q))))    (Inference Rule [] ==> '((p->(q->r))->((p->q)->(p->r)))')
#  24) ((r->q)->(r->(p&q)))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 17,23)
#  25) (r->(p&q))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 5,24)
#  26) (r->((p&q)->(r&(p&q))))    (Inference Rule [] ==> '(p->(q->(p&q)))')
#  27) ((r->((p&q)->(r&(p&q))))->(r->(r->((p&q)->(r&(p&q))))))    (Inference Rule [] ==> '(q->(p->q))')
#  28) (r->(r->((p&q)->(r&(p&q)))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 26,27)
#  29) ((r->(r->((p&q)->(r&(p&q)))))->((r->r)->(r->((p&q)->(r&(p&q))))))    (Inference Rule [] ==> '((p->(q->r))->((p->q)->(p->r)))')
#  30) ((r->r)->(r->((p&q)->(r&(p&q)))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 28,29)
#  31) (r->((p&q)->(r&(p&q))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 21,30)
#  32) ((r->((p&q)->(r&(p&q))))->((r->(p&q))->(r->(r&(p&q)))))    (Inference Rule [] ==> '((p->(q->r))->((p->q)->(p->r)))')
#  33) ((r->(p&q))->(r->(r&(p&q))))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 26,32)
#  34) (r->(r&(p&q)))    (Inference Rule ['p', '(p->q)'] ==> 'q' on lines 20,33)
# QED""")))



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: logic_utils.py

"""Python infrastructure for the Mathematical Logic through Programming book."""

from functools import lru_cache, wraps
from typing import Any, Callable, Dict, Iterator, Set, Type, TypeVar, cast

T = TypeVar('T')

def frozen(cls: Type[T]) -> Type[T]:
    """A class decorator that disallows assignment to instance variables after
    construction.

    Parameters:
        cls: class to modify.

    Returns:
        The given class, modified so that assignment to instance variable is
        disallowed after construction.
    """
    original_init = cls.__init__
    original_setattr = cls.__setattr__
    original_delattr = cls.__delattr__
    mutable_ids: Set[int] = set()
    @wraps(cls.__setattr__)
    def setattr_wrapper(self, name, value):
        if id(self) in mutable_ids:
            original_setattr(self, name, value)
        else:
            raise Exception("Cannot assign to field '" + name +
                            "' of immutable class '" + cls.__name__ + "'")
    @wraps(cls.__delattr__)
    def delattr_wrapper(self, name, value):
        if id(self) in mutable_ids:
            original_delattr(self, name, value)
        else:
            raise Exception("Cannot delete field '" + name +
                            "' of immutable class '" + cls.__name__ + "'")
    @wraps(cls.__init__)
    def init_wrapper(self, *args, **kwargs):
        mutable_ids.add(id(self))
        original_init(self, *args, **kwargs)
        mutable_ids.remove(id(self))

    setattr(cls, '__setattr__', setattr_wrapper)
    setattr(cls, '__delattr__',  delattr_wrapper)
    setattr(cls, '__init__', init_wrapper)
    return cls

class frozendict(Dict[Any, Any]):
    """An immutable variant of the built-in `dict` class."""

    def __init__(self, *args, **kwargs):
        super().update(dict(*args, **kwargs))
        
    def update(self, *args, **kwargs):
        raise Exception('Cannot modify a frozendict')

    __delattr__ = __delitem__ = __setattr__ = __setitem__ = clear = pop = \
                  popitem = setdefault =  cast(Callable[..., Any], update)
S = TypeVar('S')

def memoized_parameterless_method(method: Callable[[T], S]) -> Callable[[T], S]:
    """A method decorator for parameterless methods of immutable classes that
    memoizes the return value to avoid recalculation.

    Parameters:
        method: method to modify.

    Returns:
        The given method, modified so that after its first execution, its
        functionality is replaced with simply returning the value calculated by
        its first execution. If the value calculated by the given method has a
        `copy`\ ``()`` method, then instead of returning this value, each
        execution of the returned method, including the first one, makes a fresh
        call to this `copy`\ ``()`` method and returns the result.
    """
    methodname = method.__name__
    @wraps(method)
    def wrapper(obj):
        value = method(obj)
        if hasattr(value, 'copy'):
            new_wrapper = lambda:value.copy()
        else:
            new_wrapper = lambda:value
        object.__setattr__(obj, methodname, wraps(method)(new_wrapper))
        return new_wrapper()
    return wrapper


class __prefix_with_index_sequence_generator:
    """ A generator for a sequence of the form 'z1', 'z2', 'z3', ..., where the
    prefix 'z' is customizable. """

    def __init__(self, prefix: str) -> None:
        self.__prefix = prefix
        self.__counter = 0

    def __iter__(self) -> Iterator[str]:
        return self

    def __next__(self) -> str:
        self.__counter = self.__counter + 1
        return self.__prefix + str(self.__counter)

    def _reset_for_test(self) -> None:
        """ Reset this generator. For use by tests only """
        self.__counter = 0

#: A generator for fresh variable names. The first call to
#: `next`\ ``(``\ `fresh_variable_name_generator`\ ``)`` will return ``'z1'``,
#: the second call to `next`\ ``(``\ `fresh_variable_name_generator`\ ``)`` will
#: return ``'z2'``, and so on.
fresh_variable_name_generator: Iterator[str] = \
    __prefix_with_index_sequence_generator('z')

#: A generator for fresh constant names. The first call to
#: `next`\ ``(``\ `fresh_constant_name_generator`\ ``)`` will return ``'e1'``,
#: the second call to `next`\ ``(``\ `fresh_constant_name_generator`\ ``)`` will
#: return ``'e2'``, and so on.
fresh_constant_name_generator: Iterator[str] = \
    __prefix_with_index_sequence_generator('e')

@lru_cache(maxsize=100) # Cache the return value of is_z_and_number
def is_z_and_number(string: str) -> bool:
    """Checks if the given string is ``z`` followed by a number.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is ``z`` followed by a number, ``False``
        otherwise.
    """
    return  string[0] == 'z' and string[1:].isdecimal()



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/syntax.py

"""Syntactic handling of propositional formulas."""

from __future__ import annotations
from functools import lru_cache
from typing import Mapping, Optional, Set, Tuple, Union
from itertools import takewhile
from mltp.logic_utils import frozen, memoized_parameterless_method

@lru_cache(maxsize=100) # Cache the return value of is_variable
def is_variable(string: str) -> bool:
    """Checks if the given string is a variable name.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a variable name, ``False`` otherwise.
    """
    if type(string) == str:
        return string[0] >= 'p' and string[0] <= 'z' and (len(string) == 1 or string[1:].isdecimal())
    else:
        return False

@lru_cache(maxsize=100) # Cache the return value of is_constant
def is_constant(string: str) -> bool:
    """Checks if the given string is a constant.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a constant, ``False`` otherwise.
    """
    return string == 'T' or string == 'F'

@lru_cache(maxsize=100) # Cache the return value of is_unary
def is_unary(string: str) -> bool:
    """Checks if the given string is a unary operator.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a unary operator, ``False`` otherwise.
    """
    return string == '~'

@lru_cache(maxsize=100) # Cache the return value of is_binary
def is_binary(string: str) -> bool:
    """Checks if the given string is a binary operator.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a binary operator, ``False`` otherwise.
    """
    #return string == '&' or string == '|' or string == '->'
    # For Chapter 3:
    return string in {'&', '|',  '->', '+', '<->', '-&', '-|'}

@lru_cache(maxsize=100) # Cache the return value of split_str
def split_str(string): #returns any operator, constant, (, ), or variable if it's the first character.
    if string and string[0] == '-':
        prefix, suffix = string[:2], string[2:]
    elif string and string[0] == '<':
        prefix, suffix = string[:3], string[3:]
    else:
        prefix, suffix = string[:1], string[1:]

    if prefix and is_variable(prefix):
        while suffix and suffix[0].isnumeric():
            prefix += suffix[0]
            suffix = suffix.replace(suffix[0], '', 1)
    return prefix, suffix

@frozen
class Formula:
    """An immutable propositional formula in tree representation, composed from
    variable names, and operators applied to them.

    Attributes:
        root (`str`): the constant, variable name, or operator at the root of
            the formula tree.
        first (`~typing.Optional`\\[`Formula`]): the first operand of the root,
            if the root is a unary or binary operator.
        second (`~typing.Optional`\\[`Formula`]): the second operand of the
            root, if the root is a binary operator.
    """
    root: str
    first: Optional[Formula]
    second: Optional[Formula]

    def __init__(self, root: str, first: Optional[Formula] = None, second: Optional[Formula] = None):
        """Initializes a `Formula` from its root and root operands.

        Parameters:
            root: the root for the formula tree.
            first: the first operand for the root, if the root is a unary or
                binary operator.
            second: the second operand for the root, if the root is a binary
                operator.
        """
        if is_variable(root) or is_constant(root):
            assert first is None and second is None
            self.root = root
        elif is_unary(root):
            assert first is not None and second is None
#             self.root, self.first = root, first
            self.root = root
            self.first = first if isinstance(first, Formula) else Formula(first)
        else:
            assert is_binary(root)
            assert first is not None and second is not None
#             self.root, self.first, self.second = root, first, second
            self.root = root
            self.first = first if isinstance(first, Formula) else Formula(first)
            self.second = second if isinstance(second, Formula) else Formula(second)

    @memoized_parameterless_method
    def __repr__(self) -> str:
        """Computes the string representation of the current formula.

        Returns:
            The standard string representation of the current formula.
        """
        if is_constant(self.root) or is_variable(self.root):
            string = self.root

        elif is_unary(self.root):
            string = self.root + self.first.__repr__()

        else:
            assert is_binary(self.root)
            string = '(' + self.first.__repr__() + self.root + self.second.__repr__() + ')'
        return string
        # Task 1.1

    def __eq__(self, other: object) -> bool:
        """Compares the current formula with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is a `Formula` object that equals the
            current formula, ``False`` otherwise.
        """
        return isinstance(other, Formula) and str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        """Compares the current formula with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is not a `Formula` object or does not
            equal the current formula, ``False`` otherwise.
        """
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    @memoized_parameterless_method
    def variables(self) -> Set[str]:
        """Finds all variable names in the current formula.

        Returns:
            A set of all variable names used in the current formula.
        """
        if is_variable(self.root):
            return {self.root}

        elif is_unary(self.root):
            if is_variable(self.first):
                first = {self.first}
            else:
                first = self.first.variables()

            return first

        elif is_binary(self.root):
            if is_variable(self.first):
                first = {self.first}
            elif not is_constant(self.first):
                first = self.first.variables()

            if is_variable(self.second):
                second = {self.second}
            elif not is_constant(self.second):
                second = self.second.variables()

            return first | second

        else:
            return set()
        # Task 1.2

    @memoized_parameterless_method
    def operators(self) -> Set[str]:
        """Finds all operators in the current formula.

        Returns:
            A set of all operators (including ``'T'`` and ``'F'``) used in the
            current formula.
        """
        operators = set()
        if not is_variable(self.root):
            operators = operators | {self.root}

        if is_unary(self.root):
            return operators | self.first.operators()

        elif is_binary(self.root):
            return operators | self.first.operators() | self.second.operators()

        return operators

        # Task 1.3

    @staticmethod
    def _parse_prefix(string: str) -> Tuple[Union[Formula, None], str]:
        """Parses a prefix of the given string into a formula.

        Parameters:
            string: string to parse.

        Returns:
            A pair of the parsed formula and the unparsed suffix of the string.
            If the given string has as a prefix a variable name (e.g.,
            ``'x12'``) or a unary operator followed by a variable name, then the
            parsed prefix will include that entire variable name (and not just a
            part of it, such as ``'x1'``). If no prefix of the given string is a
            valid standard string representation of a formula then returned pair
            should be of ``None`` and an error message, where the error message
            is a string with some human-readable content.
        """
        prefix, suffix = split_str(string)
        if prefix and (is_constant(prefix) or is_variable(prefix)):
            return Formula(prefix), suffix

        elif is_unary(prefix):
            root = prefix
            first, suffix = Formula._parse_prefix(suffix)
            if first is None:
                return first, suffix
            else:
                return Formula(root, first), suffix

        elif prefix == '(':
            first, suffix = Formula._parse_prefix(suffix)
            if first is None:
                return first, suffix

            root, suffix = split_str(suffix)
            if not is_binary(root):
                return None, 'Unexpected symbol {} in {}'

            second, tail = Formula._parse_prefix(suffix)
            if second is None:
                return second, tail

            if not tail.startswith(')'):
                return None, 'Unexpected symbol {} in {}'.format(tail[:1], tail)

            return Formula(root, first, second), tail[1:]

        else:
            return None, "Unexpected input {}".format(string)
        # Task 1.4


    @staticmethod
    def is_formula(string: str) -> bool:
        """Checks if the given string is a valid representation of a formula.

        Parameters:
            string: string to check.

        Returns:
            ``True`` if the given string is a valid standard string
            representation of a formula, ``False`` otherwise.
        """
        prefix, suffix = Formula._parse_prefix(string)
        while suffix:
            try:
                prefix, suffix = Formula._parse_prefix(suffix)
            except:
                return False
            if suffix.split(' ')[0] == 'Unexpected':
                return False
        return True
        # Task 1.5

    @staticmethod
    def parse(string: str) -> Formula:
        """Parses the given valid string representation into a formula.

        Parameters:
            string: string to parse.

        Returns:
            A formula whose standard string representation is the given string.
        """
        assert Formula.is_formula(string)
        return Formula._parse_prefix(string)[0]
        # Task 1.6

    def polish(self) -> str:
        """Computes the polish notation representation of the current formula.

        Returns:
            The polish notation representation of the current formula.
        """

        if is_unary(self.root):
            first = self.first
            if isinstance(self.first, Formula):
                first = Formula.polish(first)

            return self.root + str(first)

        elif is_binary(self.root):
            first = self.first
            if isinstance(self.first, Formula):
                first = Formula.polish(first)

            second = self.second
            if isinstance(self.second, Formula):
                second = Formula.polish(second)

            return self.root + str(first) + str(second)

        else:
            return self.root
        # Optional Task 1.7

    @staticmethod
    def _parse_polish_prefix(string: str) -> Tuple[Union[Formula, None], str]:
        """Parses a prefix of the given string in polish notation into a formula.

        Parameters:
            string: string to parse.

        Returns:
            A pair containing a formula whose polish notation representation is
            the given string and the unparsed suffix of the string.
        """
        prefix, suffix = split_str(string)

        if is_variable(prefix) or is_constant(prefix):
            return Formula(prefix), suffix
        elif suffix:
            first, suffix = Formula._parse_polish_prefix(suffix)

            if is_unary(prefix):
                return Formula(prefix, first), suffix

            elif is_binary(prefix):
                second, suffix = Formula._parse_polish_prefix(suffix)
                return Formula(prefix, first, second), suffix
        else:
            return prefix, suffix
        # Optional Task 1.8

    @staticmethod
    def parse_polish(string: str) -> Formula:
        """Parses the given polish notation representation into a formula.

        Parameters:
            string: string to parse.

        Returns:
            A formula whose polish notation representation is the given string.
        """
        return Formula._parse_polish_prefix(string)[0]

    def substitute_variables(self, substitution_map: Mapping[str, Formula]) -> Formula:
        """Substitutes in the current formula, each variable name `v` that is a
        key in `substitution_map` with the formula `substitution_map[v]`.

        Parameters:
            substitution_map: mapping defining the substitutions to be
                performed.

        Returns:
            The formula resulting from performing all substitutions. Only
            variable name occurrences originating in the current formula are
            substituted (i.e., variable name occurrences originating in one of
            the specified substitutions are not subjected to additional
            substitutions).

        Examples:
            >>> Formula.parse('((p->p)|r)').substitute_variables(
            ...     {'p': Formula.parse('(q&r)'), 'r': Formula.parse('p')})
            (((q&r)->(q&r))|p)
        """
        for variable in substitution_map:
            if isinstance(substitution_map[variable], str):
                substitution_map[variable] = Formula.parse(substitution_map[variable])
            assert is_variable(variable)
            if self == Formula(variable):
                return substitution_map[variable]
            elif is_unary(self.root):
                return Formula(self.root, Formula.substitute_variables(self.first, substitution_map))
            elif is_binary(self.root):
                return Formula(self.root, Formula.substitute_variables(self.first, substitution_map), Formula.substitute_variables(self.second, substitution_map))
        return self
        # Task 3.3

    def substitute_operators(self, substitution_map: Mapping[str, Formula]) -> Formula:
        """Substitutes in the current formula, each constant or operator `op`
        that is a key in `substitution_map` with the formula
        `substitution_map[op]` applied to its (zero or one or two) operands,
        where the first operand is used for every occurrence of ``'p'`` in the
        formula and the second for every occurrence of ``'q'``.

        Parameters:
            substitution_map: mapping defining the substitutions to be
                performed.

        Returns:
            The formula resulting from performing all substitutions. Only
            operator occurrences originating in the current formula are
            substituted (i.e., operator occurrences originating in one of the
            specified substitutions are not subjected to additional
            substitutions).

        Examples:
            >>> Formula.parse('((x&y)&~z)').substitute_operators(
            ...     {'&': Formula.parse('~(~p|~q)')})
            ~(~~(~x|~y)|~~z)
        """
        for operator in substitution_map:
            assert is_constant(operator) or is_unary(operator) or is_binary(operator)
            assert substitution_map[operator].variables().issubset({'p', 'q'})

        substitution_map = {operator:substitution_map[operator] for operator in substitution_map if operator in Formula.operators(self)}
        if substitution_map:
            if is_constant(str(self)):
                return substitution_map[str(self)]

            else:
                if is_unary(self.root):
                    if self.root in substitution_map:
                        schema = substitution_map[self.root]
                        first = Formula.substitute_operators(self.first, substitution_map)
                        self = Formula.substitute_variables(schema, {'p':first})
                    else:
                        self = Formula(self.root, Formula.substitute_operators(self.first, substitution_map))

                elif is_binary(self.root):
                    if self.root in substitution_map:
                        schema = substitution_map[self.root]
                        first = Formula.substitute_operators(self.first, substitution_map)
                        second = Formula.substitute_operators(self.second, substitution_map)
                        self = Formula.substitute_variables(schema, {'p':first, 'q':second})
                    else:
                        self = Formula(self.root, Formula.substitute_operators(self.first, substitution_map), Formula.substitute_operators(self.second, substitution_map))

        return self
        # Task 3.4

    def _frege_helper(self: Formula) -> str:
        """Computes a representation of the current formula, which contains only
       the operators `~` and `->`, in Frege's notation, lacking a turnstyle.

        Returns:
            The Fregean notation representation of the current formula, lacking
            a turnstyle.
        """
        # ─ ⊢ │ ├ ┤ ┬ ┴ ┌ ┐ └ ┘ ╞ ╡
        root = self.root
        if is_binary(root):
            first = self.first._frege_helper()
            second = self.second._frege_helper()
            diff = len(first[0])-len(second[0])

            if diff > 0:
                second[0] = '─'*diff+second[0]
            second[0] = '─┬'+second[0]
            if len(second) > 1:
                for i in range(1, len(second)):
                    if diff > 0:
                        second[i] = ' '*diff+second[i]
                    second[i] = ' │'+second[i]

            if diff < 0:
                first[0] = '─'*(diff*-1)+first[0]
            first[0] = ' └'+first[0]
            if len(first) > 1:
                for i in range(1, len(first)):
                    if diff < 0:
                        first[i] = ' '*(diff*-1)+first[i]
                    first[i] = '  '+first[i]
            return [*second, *first]

        elif is_unary(root):
            first = self.first._frege_helper()
            if len(first) > 1:
                for i in range(1, len(first)):
                    first[i] = '  '+first[i]
            first[0] = '─┬'+first[0]
            return first

        else:
            assert is_variable(root) or is_constant(root)
            return ['─ '+str(self)]

    def frege(self: Formula) -> str:
        """Computes a representation of the current formula in Frege's notation.

        Returns:
            A Fregean notation representation of the current formula.
        """
        sub_map = { 'T': Formula.parse('(p->p)'),
                'F': Formula.parse('~(p->p)'),
                '&': Formula.parse('~(p->~q)'),
                '|': Formula.parse('(~p->q)'),
                '+': Formula.parse('((p->q)->~(q->p))'),
                '-|': Formula.parse('~(~p->q)'),
                '-&': Formula.parse('(p->~q)'),
                '<->': Formula.parse('~((p->q)->~(q->p))')}
        formula = Formula.substitute_operators(self, sub_map)
        string = formula._frege_helper()
        print('├─'+'\n  '.join(string))
    # Personal task

def negate_formula(formula: Formula) -> Formula:
    if is_variable(formula.root) or is_binary(formula.root):
        return Formula('~', formula)
    elif is_constant(formula.root):
        if formula == Formula.parse('T'):
            return Formula.parse('F')
        else:
            return Formula.parse('T')
    else:
        return formula.first
    # Personal task



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/prover.py

"""Axiomatic schemas of Predicate Logic, and useful proof creation maneuvers
using them."""

from typing import AbstractSet, Collection, FrozenSet, List, Mapping, \
                   Sequence, Tuple, Union

from mltp.logic_utils import fresh_variable_name_generator, is_z_and_number
from .syntax import ForbiddenVariableError, Term, Formula
from .proofs import *

class Prover:
    """A class for gradually creating a predicate-logic proof from given
    assumptions as well as from the six axioms (`AXIOMS`) Universal
    Instantiation (`UI`), Existential Introduction (`EI`), Universal
    Simplification (`US`), Existential Simplification (`ES`), Reflexivity
    (`RX`), and Meaning of Equality (`ME`).

    Attributes:
        _assumptions (`~typing.FrozenSet`\\[`~predicates.proofs.Schema`]): the
            assumptions/axioms of the proof being created, which include
            `AXIOMS`.
        _lines (`~typing.List`\\[`~predicates.proofs.Proof.Line`]): the lines so
            far of the proof being created.
        _print_as_proof_forms (`bool`): flag specifying whether the proof being
            created is being printed in real time as it forms.
    """
    _assumptions: FrozenSet[Schema]
    _lines: List[Proof.Line]
    _print_as_proof_forms: bool

    #: Axiom schema of universal instantiation
    UI = Schema(Formula.parse('(Ax[R(x)]->R(c))'), {'R', 'x', 'c'})
    #: Axiom schema of existential introduction
    EI = Schema(Formula.parse('(R(c)->Ex[R(x)])'), {'R', 'x', 'c'})
    #: Axiom schema of universal simplification
    US = Schema(Formula.parse('(Ax[(Q()->R(x))]->(Q()->Ax[R(x)]))'),
                {'Q', 'R', 'x'})
    #: Axiom schema of existential simplification
    ES = Schema(Formula.parse('((Ax[(R(x)->Q())]&Ex[R(x)])->Q())'),
                {'Q', 'R', 'x'})
    #: Axiom schema of reflexivity
    RX = Schema(Formula.parse('c=c'), {'c'})
    #: Axiom schema of meaning of equality
    ME = Schema(Formula.parse('(c=d->(R(c)->R(d)))'), {'R', 'c', 'd'})
    #: Axiomatic system for Predicate Logic, consisting of `UI`, `EI`, `US`,
    #: `ES`, `RX`, and `ME`.
    AXIOMS = frozenset({UI, EI, US, ES, RX, ME})

    def __init__(self, assumptions: Collection[Union[Schema, Formula, str]],
                 print_as_proof_forms: bool=False):
        """Initializes a `Prover` from its assumptions/additional axioms. The
        proof created by the prover initially has no lines.

        Parameters:
            assumptions: the assumptions/axioms beyond `AXIOMS` for the proof
                to be created, each specified as either a schema, a formula that
                constitutes the unique instance of the assumption, or the string
                representation of the unique instance of the assumption.
            print_as_proof_forms: flag specifying whether the proof to be
                created is to be printed in real time as it forms.
        """
        self._assumptions = \
            Prover.AXIOMS.union(
                {assumption if isinstance(assumption, Schema)
                 else Schema(assumption) if isinstance(assumption, Formula)
                 else Schema(Formula.parse(assumption))
                 for assumption in assumptions})
        self._lines = []
        self._print_as_proof_forms = print_as_proof_forms
        if self._print_as_proof_forms:
            print('Proving from assumptions/axioms:\n'
                  '  AXIOMS')
            for assumption in self._assumptions - Prover.AXIOMS:
                  print('  ' + str(assumption))
            print('Lines:')

    def qed(self) -> Proof:
        """Concludes the proof created by the current prover.

        Returns:
            A valid proof, from the assumptions of the current prover as well as
            `AXIOMS`', of the formula justified by the last line appended to the
            current prover.
        """
        conclusion = self._lines[-1].formula
        if self._print_as_proof_forms:
            print('Conclusion:', str(conclusion) + '. QED\n')
        return Proof(self._assumptions, conclusion, self._lines)

    def _add_line(self, line: Proof.Line) -> int:
        """Appends to the proof being created by the current prover the given
        validly justified line.

        Parameters:
            line: proof line that is validly justified when appended to the
                lines of the proof being created by the current prover.

        Returns:
            The line number of the appended line in the proof being created by
            the current prover.
        """
        line_number = len(self._lines)
        self._lines.append(line)
        assert line.is_valid(self._assumptions, self._lines, line_number), print(self._assumptions, self._lines, line_number)
        if self._print_as_proof_forms:
            print(('%3d) ' % line_number) + str(line.formula))
        return line_number

    def add_instantiated_assumption(self, instance: Union[Formula, str],
                                    assumption: Schema,
                                    instantiation_map: InstantiationMap) -> \
            int:
        """Appends to the proof being created by the current prover a line that
        validly justifies the given instance of the given assumptions/axioms of
        the current prover.

        Parameters:
            instance: instance to be appended, specified as either a formula or
                its string representation.
            assumption: assumption/axiom of the current prover that instantiates
                the given instance.
            instantiation_map: mapping instantiating the given instance from the
                given assumption/axiom. Each value of this map may also be given
                as a string representation (instead of a term or a formula).

        Returns:
            The line number of the newly appended line that justifies the given
            instance in the proof being created by the current prover.
        """
        if isinstance(instance, str):
            instance = Formula.parse(instance)
        instantiation_map = dict(instantiation_map)
        for key in instantiation_map:
            value = instantiation_map[key]
            if is_variable(key):
                assert is_variable(value)
            elif is_constant(key):
                if isinstance(value, str):
                    instantiation_map[key] = Term.parse(value)
                else:
                    assert isinstance(value, Term)
            else:
                assert is_relation(key)
                if isinstance(value, str):
                    instantiation_map[key] = Formula.parse(value)
                else:
                    assert isinstance(value, Formula)

        return self._add_line(Proof.AssumptionLine(instance, assumption,
                                                   instantiation_map))

    def add_assumption(self, unique_instance: Union[Formula, str]) -> int:
        """Appends to the proof being created by the current prover a line that
        validly justifies the unique instance of one of the assumptions/axioms
        of the current prover.

        Parameters:
            unique_instance: unique instance of one of the assumptions/axioms
                of the current prover, to be appended, specified as either a
                formula or its string representation.

        Returns:
            The line number of the newly appended line that justifies the given
            instance in the proof being created by the current prover.
        """
        if isinstance(unique_instance, str):
            unique_instance = Formula.parse(unique_instance)
        return self.add_instantiated_assumption(unique_instance,
                                                Schema(unique_instance), {})

    def add_tautology(self, tautology: Union[Formula, str]) -> int:
        """Appends to the proof being created by the current prover a line that
        validly justifies the given tautology.

        Parameters:
            tautology: tautology to be appended, specified as either a formula
                or its string representation.

        Returns:
            The line number of the newly appended line that justifies the given
            tautology in the proof being created by the current prover.
        """
        if isinstance(tautology, str):
            tautology = Formula.parse(tautology)
        return self._add_line(Proof.TautologyLine(tautology))

    def add_mp(self, consequent: Union[Formula, str],
               antecedent_line_number: int, conditional_line_number: int) -> \
            int:
        """Appends to the proof being created by the current prover a line that
        validly justifies the given consequent of an MP inference from the
        specified already existing lines of the proof.

        Parameters:
            consequent: consequent of MP inference to be appended, specified as
                either a formula or its string representation.
            antecedent_line_number: line number in the proof of the antecedent
                of the MP inference that derives the given formula.
            conditional_line_number: line number in the proof of the conditional
                of the MP inference that derives the given formula.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.
        """
        if isinstance(consequent, str):
            consequent = Formula.parse(consequent)
        return self._add_line(Proof.MPLine(consequent, antecedent_line_number,
                                           conditional_line_number))

    def add_ug(self, quantified: Union[Formula, str],
               nonquantified_line_number: int) -> int:
        """Appends to the proof being created by the current prover a line that
        validly justifies the given universally quantified formula, the
        statement quantified by which is the specified already existing line of
        the proof.

        Parameters:
            quantified: universally quantified formula to be appended, specified
                as either a formula or its string representation.
            nonquantified_line_number: line number in the proof of the statement
                quantified by the given formula.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.
        """
        if isinstance(quantified, str):
            quantified = Formula.parse(quantified)
        return self._add_line(Proof.UGLine(quantified,
                                           nonquantified_line_number))

    def add_and_introduction(self, firstline, secondline):
        first = self._lines[firstline].formula
        second = self._lines[secondline].formula
        tautology = Formula('->',first,
                        Formula('->',second,
                            Formula('&', first, second)))
        step1 = self.add_tautology(tautology)
        step2 = self.add_mp(tautology.second, firstline, step1)
        return self.add_mp(tautology.second.second, secondline, step2)
        # personal task

    def add_and_elimination(self, conjunction_line, desired_operand):
        conjunction = self._lines[conjunction_line].formula
        assert conjunction.root == '&'
        assert conjunction.first == desired_operand or conjunction.second == desired_operand
        tautology = Formula('->', conjunction, desired_operand)
        step1 = self.add_tautology(tautology)
        return self.add_mp(desired_operand, conjunction_line, step1)
        # personal task

    def add_or_introduction(self, conjunct1, conjunct1line, conjunct2, conjunct2line):
        pass

    def add_or_elimination(self, ):
        pass

    def add_proof(self, conclusion: Union[Formula, str], proof: Proof) -> int:
        """Appends to the proof being created by the current prover a validly
        justified inlined version of the given proof of the given conclusion,
        the last line of which validly justifies the given formula.

        Parameters:
            conclusion: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            proof: valid proof of the given formula from a subset of the
                assumptions/axioms of the current prover.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.
        """
        if isinstance(conclusion, str):
            conclusion = Formula.parse(conclusion)
        assert proof.conclusion == conclusion
        assert proof.assumptions.issubset(self._assumptions)
        line_shift = len(self._lines)
        for line in proof.lines:
            if type(line) in {Proof.AssumptionLine, Proof.TautologyLine}:
                self._add_line(line)
            elif isinstance(line, Proof.MPLine):
                self.add_mp(line.formula,
                            line.antecedent_line_number + line_shift,
                            line.conditional_line_number + line_shift)
            else:
                assert isinstance(line, Proof.UGLine)
                self.add_ug(line.formula,
                            line.nonquantified_line_number + line_shift)
        line_number = len(self._lines) - 1
        assert self._lines[line_number].formula == conclusion
        return line_number

    def add_universal_instantiation(self, instantiation: Union[Formula, str],
                                    line_number: int, term: Union[Term, str]) \
            -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given formula, which is the result of substituting a term for the
        outermost universally quantified variable name in the formula in the
        specified already existing line of the proof.

        Parameters:
            instantiation: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_number: line number in the proof of a universally quantified
                formula of the form `Ax[statement]`.
            term: term, specified as either a term or its string representation,
                that when substituted into the free occurrences of `x` in
                `statement` yields the given formula.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.

        Examples:
            If Line `line_number` contains the formula
            `'Ay[Az[f(x,y)=g(z,y)]]'` and `term` is `'h(w)'`, then
            `instantiation` should be `'Az[f(x,h(w))=g(z,h(w))]'`.
        """
        if isinstance(instantiation, str):
            instantiation = Formula.parse(instantiation)
        assert line_number < len(self._lines)
        quantified = self._lines[line_number].formula
        assert quantified.root == 'A'
        if isinstance(term, str):
            term = Term.parse(term)
        assert instantiation == \
               quantified.statement.substitute({quantified.variable: term})

        instantiation_map = {'R': quantified.statement.substitute({quantified.variable: Term('_')}),
                             'c': term,
                             'x': quantified.variable}

        step1 = self.add_instantiated_assumption(Formula('->', quantified, instantiation), Prover.UI, instantiation_map)
        return self.add_mp(instantiation, line_number, step1)
        # Task 10.1

    def add_tautological_implication(self, implication: Union[Formula, str],
                                     line_numbers: AbstractSet[int]) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given formula, which is a tautological implication of the formulas in
        the specified already existing lines of the proof.

        Parameters:
            implication: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_numbers: line numbers in the proof of formulas of which
                `implication` is a tautological implication.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.
        """
        if isinstance(implication, str):
            implication = Formula.parse(implication)
        for line_number in line_numbers:
            assert line_number < len(self._lines)

        line_numbers = sorted(line_numbers)
        tautology = implication
        for line in line_numbers[::-1]:
            tautology = Formula('->', self._lines[line].formula, tautology)
        step = self.add_tautology(tautology)
        if len(line_numbers) > 1:
            for line in line_numbers[:-1]:
                tautology = tautology.second
                step = self.add_mp(tautology, line, step)
        return self.add_mp(implication, line_numbers[-1], step)
        # Task 10.2

    def add_existential_derivation(self, consequent: Union[Formula, str],
                                   line_number1: int, line_number2: int) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given formula, which is the consequent of the formula in the second
        specified already existing line of the proof, whose antecedent is
        existentially quantified by the formula in the first specified already
        existing line of the proof.

        Parameters:
            consequent: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_number1: line number in the proof of an existentially
                quantified formula of the form
                `Ex[antecedent(x)]`, where `x`
                is a variable name that may have free occurrences in
                `antecedent(x)` but has no free occurrences in `consequent`.
            line_number2: line number in the proof of the formula
                `(antecedent(x)->consequent)`.

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.
        """
        if isinstance(consequent, str):
            consequent = Formula.parse(consequent)
        assert line_number1 < len(self._lines)
        quantified = self._lines[line_number1].formula
        assert quantified.root == 'E'
        assert quantified.variable not in consequent.free_variables(), print(f'\nThe variable quantified in line1, {quantified.variable}, is in the consequent {consequent}\n')
        assert line_number2 < len(self._lines)
        conditional = self._lines[line_number2].formula
        assert conditional == Formula('->', quantified.statement, consequent), print(f'\nConditional is {conditional},\nshould be      {Formula("->", quantified.statement, consequent)}\n')

        variable = quantified.variable
        step1 = self.add_ug(Formula('A', variable, self._lines[line_number2].formula), line_number2)
        instantiation_map = {'Q': consequent,
                        'R': quantified.statement.substitute({variable:Term('_')}),
                        'x': variable}
        step2 = self.add_instantiated_assumption(Prover.ES.instantiate(instantiation_map), Prover.ES, instantiation_map)
        return self.add_tautological_implication(consequent, {line_number1, step1, step2})
        # Task 10.3

    def add_flipped_equality(self, flipped: Union[Formula, str],
                             line_number: int) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given equality, which is the result of exchanging the two sides of the
        equality in the specified already existing line of the proof.

        Parameters:
            flipped: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_number: line number in the proof of an equality that is the
                same as the given equality, except that the two sides of the
                equality are exchanged.

        Returns:
            The line number of the newly appended line that justifies the given
            equality in the proof being created by the current prover.
        """
        if isinstance(flipped, str):
            flipped = Formula.parse(flipped)
        assert is_equality(flipped.root)
        assert line_number < len(self._lines)
        equality = self._lines[line_number].formula
        assert equality == Formula('=', [flipped.arguments[1],
                                         flipped.arguments[0]])
        x, y = equality.arguments
        instantiation_map = {'R': Formula('=', [Term('_'), x]), 'c': x, 'd': y}
        step1 = self.add_instantiated_assumption(Prover.ME.instantiate(instantiation_map), Prover.ME, instantiation_map)
        step2 = self.add_mp(self._lines[step1].formula.second, line_number, step1)
        step3 = self.add_instantiated_assumption(Prover.RX.instantiate({'c':x}), Prover.RX, {'c':x})
        return self.add_mp(self._lines[step2].formula.second, step3, step2)

        # Task 10.6

    def add_free_instantiation(self, instantiation: Union[Formula, str],
                               line_number: int,
                               substitution_map:
                               Mapping[str, Union[Term, str]]) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given formula, which is the result of substituting terms for free
        variable names in the formula in the specified already existing line of
        the proof.

        Parameters:
            instantiation: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_number: line number in the proof of a formula with free
                variable names, which contains no variable names that are `z`
                followed by a number.
            substitution_map: mapping from free variable names of the formula
                with the given line number to terms that contain no variable
                names that are `z` followed by a number, to be substituted for
                them to obtain the given formula. Each value of this map may
                also be given as a string representation (instead of a term).
                Only variable name occurrences originating in the formula with
                the given line number are substituted (i.e., variable names
                originating in one of the specified substitutions are not
                subjected to additional substitutions).

        Returns:
            The line number of the newly appended line that justifies the given
            formula in the proof being created by the current prover.

        Examples:
            If Line `line_number` contains the formula
            `'(z=5&Az[f(x,y)=g(z,y)])'` and `substitution_map` is
            `{'y': 'h(w)', 'z': 'y'}`, then `instantiation` should be
            `'(y=5&Az[f(x,h(w))=g(z,h(w))])'`.
        """
        if isinstance(instantiation, str):
            instantiation = Formula.parse(instantiation)
        substitution_map = dict(substitution_map)
        for variable in substitution_map:
            assert is_variable(variable)
            term = substitution_map[variable]
            if isinstance(term, str):
                substitution_map[variable] = term = Term.parse(term)
            for variable in term.variables():
                assert not is_z_and_number(variable)
        assert line_number < len(self._lines)
        original = self._lines[line_number].formula
        assert original.free_variables().issuperset(substitution_map.keys())
        assert instantiation == original.substitute(substitution_map)

        nested = False
        for variable1 in substitution_map:
            variables = substitution_map[variable1].variables() | substitution_map[variable1].constants()
            for variable2 in substitution_map:
                if not variable1 == variable2:
                    if variable2 in variables:
                        nested = True
                        break

        if nested:
            for variable in original.variables(): #I moved this check to only complex cases where next(fresh_variable_name_generator) is actually used
                assert not is_z_and_number(variable), print('instantiation contains a z-and-number variable and is complex. I may not be able to freely instantiate without conflicts.')
            temp_variables = {}
            new_variables = {}

            #Complex case - Produce intermediary dictionaries and perform universal generalization over all keys in substitution map
            for original_variable in substitution_map:
                temp_variable = next(fresh_variable_name_generator)
                temp_variables[original_variable] = temp_variable
                new_variables[temp_variable] = substitution_map[original_variable]
                line_number = self.add_ug(Formula('A', original_variable, self._lines[line_number].formula), line_number)

            #Universal instantiation to replace all original variables with temporary variables
            for _ in substitution_map:
                quantified_variable = self._lines[line_number].formula.variable
                new_formula = self._lines[line_number].formula.statement.substitute({quantified_variable: Term(temp_variables[quantified_variable])})
                line_number = self.add_universal_instantiation(new_formula, line_number, temp_variables[quantified_variable])

            #Universal generalization over all temporary variables
            for temp_variable in new_variables:
                line_number = self.add_ug(Formula('A', temp_variable, self._lines[line_number].formula), line_number)

            #Universal instantiation to replace all temporary variables with desired variables
            for _ in new_variables:
                quantified_variable = self._lines[line_number].formula.variable
                new_formula = self._lines[line_number].formula.statement.substitute({quantified_variable: new_variables[quantified_variable]})
                line_number = self.add_universal_instantiation(new_formula, line_number, new_variables[quantified_variable])

        else:
            #Simple case - Universal instantiation over all original variables
            for variable in substitution_map:
                line_number = self.add_ug(Formula('A', variable, self._lines[line_number].formula), line_number)

            #Universal instantiation to replace them with desired variables
            for _ in substitution_map:
                quantified_variable = self._lines[line_number].formula.variable
                new_formula = self._lines[line_number].formula.statement.substitute({quantified_variable: substitution_map[quantified_variable]})
                line_number = self.add_universal_instantiation(new_formula, line_number, substitution_map[quantified_variable])
        return line_number
        # Task 10.7

    def add_substituted_equality(self, substituted: Union[Formula, str],
                                 line_number: int,
                                 parametrized_term: Union[Term, str]) -> \
            int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given equality, whose two sides are the results of substituting the two
        respective sides of the equality in the specified already existing line
        of the proof into the given parametrized term.

        Parameters:
            substituted: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation.
            line_number: line number in the proof of an equality.
            parametrized_term: term parametrized by the constant name `'_'`,
                specified as either a term or its string representation, such
                that substituting each of the two sides of the equality with the
                given line number into this parametrized term respectively
                yields each of the two sides of the given equality.

        Returns:
            The line number of the newly appended line that justifies the given
            equality in the proof being created by the current prover.

        Examples:
            If Line `line_number` contains the formula `'g(x)=h(y)'` and
            `parametrized_term` is `'_+7'`, then `substituted` should be
            `'g(x)+7=h(y)+7'`.
        """
        if isinstance(substituted, str):
            substituted = Formula.parse(substituted)
        assert is_equality(substituted.root)
        assert line_number < len(self._lines)
        equality = self._lines[line_number].formula
        assert is_equality(equality.root)
        if isinstance(parametrized_term, str):
            parametrized_term = Term.parse(parametrized_term)
        assert substituted == \
               Formula('=', [parametrized_term.substitute(
                                 {'_': equality.arguments[0]}),
                             parametrized_term.substitute(
                                 {'_': equality.arguments[1]})])
        x, y = equality.arguments
        nonparametrized_term = parametrized_term.substitute({'_': x})
        phi = Formula('=', [nonparametrized_term, parametrized_term])
        instantiation_map = {'R': phi, 'c': x, 'd': y}
        me = Prover.ME.instantiate(instantiation_map)
        step1 = self.add_instantiated_assumption(me, Prover.ME, instantiation_map)
        step2 = self.add_mp(me.second, line_number, step1)
        step3 = self.add_instantiated_assumption(Prover.RX.instantiate({'c':nonparametrized_term}), Prover.RX, {'c':nonparametrized_term})
        return self.add_mp(me.second.second, step3, step2)
        # Task 10.8

    def _add_chaining_of_two_equalities(self, line_number1: int,
                                        line_number2: int) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies an
        equality that is the result of chaining together the two equalities in
        the specified already existing lines of the proof.

        Parameters:
            line_number1: line number in the proof of an equality of the form
                `'`first`=`second`'`.
            line_number2: line number in the proof of an equality of the form
                `'`second`=`third`'`.

        Returns:
            The line number of the newly appended line that justifies the
            equality `'`first`=`third`'` in the proof being
            created by the current prover.

        Examples:
            If Line `line_number1` contains the formula `'a=b'` and Line
            `line_number2` contains the formula `'b=f(b)'`, then the last
            appended line will contain the formula `'a=f(b)'`.
        """
        assert line_number1 < len(self._lines)
        equality1 = self._lines[line_number1].formula
        assert is_equality(equality1.root)
        assert line_number2 < len(self._lines)
        equality2 = self._lines[line_number2].formula
        assert is_equality(equality2.root)
        assert equality1.arguments[1] == equality2.arguments[0]

        x, y = equality1.arguments
        z = equality2.arguments[1]
        inst_map = {'R': Formula('=',[x,Term('_')]), 'c': y, 'd': z}
        me = Prover.ME.instantiate(inst_map)
        step1 = self.add_instantiated_assumption(me, Prover.ME, inst_map)
        step2 = self.add_mp(me.second, line_number2, step1)
        return self.add_mp(me.second.second, line_number1, step2)
        # Task 10.9a

    def add_chained_equality(self, chained: Union[Formula, str],
                             line_numbers: Sequence[int]) -> int:
        """Appends to the proof being created by the current prover a sequence
        of validly justified lines, the last of which validly justifies the
        given equality, which is the result of chaining together the equalities
        in the specified already existing lines of the proof.

        Parameters:
            chained: conclusion of the sequence of lines to be appended,
                specified as either a formula or its string representation,
                of the form `'`first`=`last`'`.
            line_numbers: line numbers in the proof of equalities of the form
                `'`first`=`second`'`,
                `'`second`=`third`'`, ...,
                `'`before_last`=`last`'`, i.e., the left-hand
                side of the first equality is the left-hand side of the given
                equality, the right-hand of each equality (except for the last)
                is the left-hand side of the next equality, and the right-hand
                side of the last equality is the right-hand side of the given
                equality.

        Returns:
            The line number of the newly appended line that justifies the given
            equality in the proof being created by the current prover.

            Examples:
            If `line_numbers` is `[7,3,9]`, Line 7 contains the formula
            `'a=b'`, Line 3 contains the formula `'b=f(b)'`, and Line 9
            contains the formula `'f(b)=0'`, then `chained` should be
            `'a=0'`.
        """
        if isinstance(chained, str):
            chained = Formula.parse(chained)
        assert is_equality(chained.root)
        assert len(line_numbers) >= 2
        current_term = chained.arguments[0]
        for line_number in line_numbers:
            assert line_number < len(self._lines)
            equality = self._lines[line_number].formula
            assert is_equality(equality.root)
            assert equality.arguments[0] == current_term
            current_term = equality.arguments[1]
        assert chained.arguments[1] == current_term

        step0 = line_numbers[0]
        equality1 = self._lines[step0].formula

        for line in line_numbers[1:]:
            equality2 = self._lines[line].formula
            x, y = equality1.arguments
            z = equality2.arguments[1]
            inst_map = {'R': Formula('=', [x, Term('_')]), 'c': y, 'd': z}
            me = Prover.ME.instantiate(inst_map)
            step1 = self.add_instantiated_assumption(me, Prover.ME, inst_map)
            step2 = self.add_mp(me.second, line, step1)
            step0 = self.add_mp(me.second.second, step0, step2)
            equality1 = me.second.second
        return step0
        # Task 10.9b



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/completeness.py

"""Building blocks for proving the Completeness Theorem for Predicate Logic."""

from typing import AbstractSet, Container, Set, Union
import time
from itertools import product
from mltp.logic_utils import fresh_constant_name_generator
from .syntax import *
from .semantics import *
from .proofs import *
from .prover import *
from .deduction import *
from .prenex import *

def get_constants(formulas: AbstractSet[Formula]) -> Set[str]:
    """Finds all constant names in the given formulas.

    Parameters:
        formulas: formulas to find all constant names in.

    Returns:
        A set of all constant names used in one or more of the given formulas.
    """
    constants = set()
    for formula in formulas:
        constants.update(formula.constants())
    return constants

def get_relations(formulas: AbstractSet[Formula]) -> Set[str]:
    """Finds all relation names in the given formulas.

    Parameters:
        formulas: formulas to find all relations in.

    Returns:
        A set of all relations used in one or more of the given formulas.
    """
    relations = set()
    for formula in formulas:
        relations.update(formula.relations())
    return relations
    # personal task

def is_closed(sentences: AbstractSet[Formula]) -> bool:
    """Checks whether the given set of prenex-normal-form sentences is closed.

    Parameters:
        sentences: set of prenex-normal-form sentences to check.

    Returns:
        `True` if the given set of sentences is primitively, universally, and
        existentially closed; `False` otherwise.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0
    return is_primitively_closed(sentences) and \
           is_universally_closed(sentences) and \
           is_existentially_closed(sentences)

def is_primitively_closed(sentences: AbstractSet[Formula]) -> bool:
    """Checks whether the given set of prenex-normal-form sentences is
    primitively closed.

    Parameters:
        sentences: set of prenex-normal-form sentences to check.

    Returns:
        `True` if for every n-ary relation name from the given sentences, and
        for every n (not necessarily distinct) constant names from the given
        sentences, either the invocation of this relation name over these
        constant names (in order), or the negation of this invocation (or both),
        is one of the given sentences; `False` otherwise.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0
    relations = set.union(*[sentence.relations() for sentence in sentences])
    constant_names = sorted(set.union(*[sentence.constants() for sentence in sentences]))
    for relation in relations:
        for constants in product(constant_names, repeat=relation[1]):
            formula = Formula(relation[0], constants)
            if not (formula in sentences or Formula('~',formula) in sentences):
                return False
    return True
    # Task 12.1a

def is_universally_closed(sentences: AbstractSet[Formula]) -> bool:
    """Checks whether the given set of prenex-normal-form sentences is
    universally closed.

    Parameters:
        sentences: set of prenex-normal-form sentences to check.

    Returns:
        `True` if for every universally quantified sentence from the given set
        of sentences, and for every constant name from these sentences, the
        statement quantified by this sentence, with every free occurrence of the
        universal quantification variable name replaced with this constant name,
        is also in the given set; `False` otherwise.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0

    constant_names = sorted(set.union(*[sentence.constants() for sentence in sentences]))
    constant_names = [Term(constant) for constant in constant_names]
    for sentence in sentences:
        if sentence.root == 'A':
            for constant in constant_names:
                if not sentence.statement.substitute({sentence.variable: constant}) in sentences:
                    return False
    return True
    # Task 12.1b

def is_existentially_closed(sentences: AbstractSet[Formula]) -> bool:
    """Checks whether the given set of prenex-normal-form sentences is
    existentially closed.

    Parameters:
        sentences: set of prenex-normal-form sentences to check.

    Returns:
        `True` if for every existentially quantified sentence from the given
        set of sentences there exists a constant name such that the statement
        quantified by this sentence, with every free occurrence of the
        existential quantification variable name replaced with this constant
        name, is also in the given set; `False` otherwise.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0
    constant_names = sorted(set.union(*[sentence.constants() for sentence in sentences]))
    constant_names = [Term(constant) for constant in constant_names]
    for sentence in sentences:
        if sentence.root == 'E':
            if not any(sentence.statement.substitute({sentence.variable: constant}) in sentences for constant in constant_names):
                return False
    return True
    # Task 12.1c

def find_unsatisfied_quantifier_free_sentence(sentences: Container[Formula],
                                              model: Model[str],
                                              unsatisfied: Formula) -> Formula:
    """
    Given a universally and existentially closed set of prenex-normal-form
    sentences, given a model whose universe is the set of all constant names
    from the given sentences, and given a sentence from the given set that the
    given model does not satisfy, finds a quantifier-free sentence from the
    given set that the given model does not satisfy.

    Parameters:
        sentences: universally and existentially closed set of
            prenex-normal-form sentences, which is only to be accessed using
            containment queries, i.e., using the `in` operator as in:

            >>> if sentence in sentences:
            ...     print('contained!')

        model: model for all element names from the given sentences, whose
            universe is `get_constants`(`sentences`)`.
        unsatisfied: sentence (which possibly contains quantifiers) from the
            given sentences that is not satisfied by the given model.

    Returns:
        A quantifier-free sentence from the given set of sentences that is not
        satisfied by the given model.
    """
    # We assume that every formula in sentences is in prenex normal form and has
    # no free variable names, that sentences are universally and existentially
    # closed, and that the set of constant names that appear somewhere in
    # sentences is model.universe; but we cannot assert these since we cannot
    # iterate over sentences.
    for constant in model.universe:
        assert is_constant(constant)
    assert is_in_prenex_normal_form(unsatisfied)
    assert len(unsatisfied.free_variables()) == 0
    assert unsatisfied in sentences
    assert not model.evaluate_formula(unsatisfied)

    if is_quantifier_free(unsatisfied):
        return unsatisfied
    else:
        for constant in model.universe:
            if (u:=unsatisfied.statement.substitute({unsatisfied.variable: Term(constant)})) in sentences:
                if not model.evaluate_formula(u):
                    return find_unsatisfied_quantifier_free_sentence(sentences, model, u)
    # Task 12.2

def get_primitives(quantifier_free: Formula) -> Set[Formula]:
    """Finds all primitive subformulas of the given quantifier-free formula.

    Parameters:
        quantifier_free: quantifier-free formula that contains no function names
            and no equalities, whose subformulas are to be searched.

    Returns:
        The primitive subformulas (i.e., relation invocations) of which the
        given quantifier-free formula is composed using logical operators.

    Examples:
        The primitive subformulas of `'(R(c1,d)|~(Q(c1)->~R(c2,a)))'` are
        `'R(c1,d)'`, `'Q(c1)'`, and `'R(c2,a)'`.
    """
    assert is_quantifier_free(quantifier_free)
    assert len(quantifier_free.functions()) == 0
    assert '=' not in str(quantifier_free)
    root = quantifier_free.root
    if is_relation(root):
        return {quantifier_free}
    elif is_unary(root):
        return get_primitives(quantifier_free.first)
    else:
        return get_primitives(quantifier_free.first) | get_primitives(quantifier_free.second)
    # Task 12.3a

def model_or_inconsistency(sentences: AbstractSet[Formula]) -> \
        Union[Model[str], Proof]:
    """Either finds a model in which the given closed set of prenex-normal-form
    sentences holds, or proves a contradiction from these sentences.

    Parameters:
        sentences: closed set of prenex-normal-form sentences that contain no
            function names and no equalities, to either find a model of, or
            prove a contradiction from.

    Returns:
        A model in which all of the given sentences hold if such exists,
        otherwise a valid proof of  a contradiction from the given formulas via
        `~predicates.prover.Prover.AXIOMS`.
    """
    assert is_closed(sentences)
    for sentence in sentences:
        assert len(sentence.functions()) == 0
        assert '=' not in str(sentence)

    universe = {i for i in get_constants(sentences)}
    constant_interpretations = {str(constant):str(constant) for constant in universe}
    relations = get_relations(sentences)
    relation_interpretations = {relation[0]: set() for relation in relations}

    for sentence in sentences:
        if is_relation(root:=sentence.root):
            relation_interpretations[root].add(tuple(str(i) for i in sentence.arguments))

    model = Model(universe, constant_interpretations, relation_interpretations)

    for sentence in sentences:
        if not model.evaluate_formula(sentence):
            unsatisfied = find_unsatisfied_quantifier_free_sentence(sentences, model, sentence)
            prover = Prover(Prover.AXIOMS.union({Schema(s) for s in sentences}))
            premises = set()
            premises.add(prover.add_assumption(unsatisfied))
            for primitive in get_primitives(unsatisfied):
                if not primitive in sentences:
                    primitive = Formula('~', primitive)
                    assert primitive in sentences
                premises.add(prover.add_assumption(primitive))
            prover.add_tautological_implication('((P()-&(P()-&P()))-&(P()-&(P()-&P())))', premises)
            return prover.qed()
    return model
    # Task 12.3b

def combine_contradictions(proof_from_affirmation: Proof,
                           proof_from_negation: Proof) -> Proof:
    """Combines the given two proofs of contradictions, both from the same
    assumptions/axioms except that the latter has an extra assumption that is
    the negation of an extra assumption that the former has, into a single proof
    of a contradiction from only the common assumptions/axioms.

    Parameters:
        proof_from_affirmation: valid proof of a contradiction from one or more
            assumptions/axioms that are all sentences and that include
            `~predicates.prover.Prover.AXIOMS`.
        proof_from_negation: valid proof of a contradiction from the same
            assumptions/axioms of `proof_from_affirmation`, but with one
            simple assumption (i.e., without any templates) `assumption`
            replaced with its negation `'~`assumption`'`.

    Returns:
        A valid proof of a contradiction from only the assumptions/axioms common
        to the given proofs (i.e., without `assumption` or its negation).
    """
    assert proof_from_affirmation.is_valid()
    assert proof_from_negation.is_valid()
    common_assumptions = proof_from_affirmation.assumptions.intersection(
        proof_from_negation.assumptions)
    assert len(common_assumptions) == \
           len(proof_from_affirmation.assumptions) - 1
    assert len(common_assumptions) == len(proof_from_negation.assumptions) - 1
    affirmed_assumption = list(proof_from_affirmation.assumptions -
                               common_assumptions)[0]
    negated_assumption = list(proof_from_negation.assumptions -
                              common_assumptions)[0]
    assert len(affirmed_assumption.templates) == 0
    assert len(negated_assumption.templates) == 0
    assert negated_assumption.formula == \
           Formula('~', affirmed_assumption.formula)
    assert proof_from_affirmation.assumptions.issuperset(Prover.AXIOMS)
    assert proof_from_negation.assumptions.issuperset(Prover.AXIOMS)
    for assumption in common_assumptions.union({affirmed_assumption,
                                                negated_assumption}):
        assert len(assumption.formula.free_variables()) == 0

    proof1 = prove_by_way_of_contradiction(proof_from_affirmation, affirmed_assumption.formula)
    proof2 = prove_by_way_of_contradiction(proof_from_negation, negated_assumption.formula)

    prover = Prover(common_assumptions)
    step1 = prover.add_proof(proof1.conclusion, proof1)
    step2 = prover.add_proof(proof2.conclusion, proof2)
    prover.add_tautological_implication('(P()-|(P()-|P()))', {step1, step2})
    return prover.qed().clean()
    # Task 12.4

def eliminate_universal_instantiation_assumption(proof: Proof,
                                                 universal: Formula,
                                                 constant: str) -> Proof:
    """Converts the given proof of a contradiction, whose assumptions/axioms
    include `universal` and `instantiation`, where the latter is the universal
    instantiation of the former with the constant name `constant`, to a proof
    of a contradiction from the same assumptions without `instantiation`.

    Parameters:
        proof: valid proof of a contradiction from one or more
            assumptions/axioms that are all sentences and that include
            `~predicates.prover.Prover.AXIOMS`.
        universal: assumption of the given proof that is universally quantified.
        constant: constant name such that the formula `instantiation` obtained
            from the statement quantified by `universal` by replacing all free
            occurrences of the universal quantification variable name by the
            given constant name, is an assumption of the given proof.

    Returns:
        A valid proof of a contradiction from the assumptions/axioms of the
        given proof except `instantiation`.
    """
    assert proof.is_valid()
    assert Schema(universal) in proof.assumptions
    assert universal.root == 'A'
    assert is_constant(constant)
    assert Schema(universal.statement.substitute({universal.variable:
                                                  Term(constant)})) in \
           proof.assumptions
    for assumption in proof.assumptions:
        assert len(assumption.formula.free_variables()) == 0
    variable = universal.variable
    statement = universal.statement
    proof1 = prove_by_way_of_contradiction(proof, statement.substitute({variable: Term(constant)}))
    prover = Prover(proof1.assumptions)
    inst_map = {'R': statement.substitute({variable: Term('_')}),
                'c': Term(constant),
                'x': variable}
    step1 = prover.add_proof(proof1.conclusion, proof1)
    step2 = prover.add_assumption(universal)
    step3 = prover.add_instantiated_assumption(Prover.UI.instantiate(inst_map), Prover.UI, inst_map)
    step4 = prover.add_mp(statement.substitute({variable: Term(constant)}), step2, step3)
    step5 = prover.add_tautological_implication('~((P()-|P())<->(P()-&P()))', {step1, step4})
    return prover.qed()
    # Task 12.5

def universal_closure_step(sentences: AbstractSet[Formula]) -> Set[Formula]:
    """Augments the given sentences with all universal instantiations of each
    universally quantified sentence from these sentences, with respect to all
    constant names from these sentences.

    Parameters:
        sentences: prenex-normal-form sentences to augment with their universal
            instantiations.

    Returns:
        A set of all of the given sentences, and in addition any formula that
        can be obtained from the statement quantified by any universally
        quantified sentence from the given sentences by replacing all
        occurrences of the quantification variable name with some constant name
        from the given sentences.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0
    constants = [Term(constant) for constant in get_constants(sentences)]
    new_sentences = set(sentences)
    for sentence in sentences:
        if sentence.root == 'A':
            for constant in constants:
                new_sentences.add(sentence.statement.substitute({sentence.variable: constant}))
    return new_sentences
    # Task 12.6

def replace_constant(proof: Proof, constant: str, variable: str = 'zz') -> \
        Proof:
    """Replaces all occurrences of the given constant name in the given proof
    with the given variable name.

    Parameters:
        proof: valid proof in which to replace.
        constant: constant name that does not appear as a template constant name
            in any of the assumptions of the given proof.
        variable: variable name that does not appear anywhere in the given proof
            or in its assumptions.

    Returns:
        A valid proof where every occurrence of the given constant name in the
        given proof and in its assumptions is replaced with the given variable
        name.
    """
    assert proof.is_valid()
    assert is_constant(constant)
    assert is_variable(variable)
    for assumption in proof.assumptions:
        assert constant not in assumption.templates
        assert variable not in assumption.formula.variables()
    for line in proof.lines:
        assert variable not in line.formula.variables()

    substitution = {constant: Term(variable)}

    assumption_dict = {}
    for assumption in proof.assumptions:
        assumption_dict[assumption] = Schema(
                                        assumption.formula.substitute(substitution),
                                        assumption.templates
                                        )
    new_assumptions = set(assumption_dict.values())

    new_lines = []
    for line in proof.lines:
        new_formula = line.formula.substitute(substitution)
        if isinstance(line, Proof.UGLine):
            new_line = Proof.UGLine(new_formula,
                             line.nonquantified_line_number)

        elif isinstance(line, Proof.MPLine):
            new_line = Proof.MPLine(new_formula,
                             line.antecedent_line_number,
                             line.conditional_line_number)

        elif isinstance(line, Proof.TautologyLine):
            new_line = Proof.TautologyLine(new_formula)

        else: # AssumptionLine
            new_inst_map = {}
            for key, value in line.instantiation_map.items():
                if is_relation(key) or is_constant(key):
                    if constant in value.constants():
                        new_inst_map[key] = value.substitute(substitution)
                    else:
                        new_inst_map[key] = value
                else:
                    new_inst_map[key] = value
            new_line = Proof.AssumptionLine(new_formula,
                                     assumption_dict[line.assumption],
                                     new_inst_map)
        new_lines.append(new_line)

    new_conclusion = new_lines[-1].formula
    return Proof(new_assumptions, new_conclusion, new_lines)
    # Task 12.7a

def eliminate_existential_witness_assumption(proof: Proof,
                                             existential: Formula,
                                             constant: str) -> Proof:
    """Converts the given proof of a contradiction, whose assumptions/axioms
    include `existential` and `witness`, where the latter is the existential
    witness of the former with the witnessing constant name `constant`, to a
    proof of a contradiction from the same assumptions without `witness`.

    Parameters:
        proof: valid proof, which does not contain the variable name `'zz'` in
            its lines or assumptions, of a contradiction from one or more
            assumptions/axioms that are all sentences and that include
            `~predicates.prover.Prover.AXIOMS`.
        existential: assumption of the given proof that is existentially
            quantified.
        constant: constant name such that the formula `witness` obtained from
            from the statement quantified by `existential` by replacing all free
            occurrences of the existential quantification variable name by the
            given constant name, is an assumption of the given proof, and such
            that this constant name does not appear in any assumption of the
            given proof except `witness`.

    Returns:
        A valid proof of a contradiction from the assumptions/axioms of the
        given proof except `witness`.
    """
    assert proof.is_valid()
    assert Schema(existential) in proof.assumptions
    assert existential.root == 'E'
    assert is_constant(constant)
    witness = existential.statement.substitute({existential.variable:
                                                Term(constant)})
    assert Schema(witness) in proof.assumptions
    for assumption in proof.assumptions:
        assert len(assumption.formula.free_variables()) == 0
        assert 'zz' not in assumption.formula.variables()
    for assumption in proof.assumptions - {Schema(witness)}:
        assert constant not in assumption.formula.constants()
    for line in proof.lines:
        assert 'zz' not in line.formula.variables()

    variable = existential.variable
    new_witness = witness.substitute({constant: Term('zz')})
    converted_proof = replace_constant(proof, constant, 'zz')
    witness_removed = prove_by_way_of_contradiction(converted_proof, new_witness)

    prover = Prover(proof.assumptions - {Schema(witness)})
    step1 = prover.add_proof(witness_removed.conclusion, witness_removed)

    sub_map = {'zz': Term(variable)}
    unquantified_formula = Formula('~', new_witness.substitute(sub_map))
    universally_quantified_formula = Formula('A', variable, unquantified_formula)
    step2 = prover.add_free_instantiation(unquantified_formula, step1, sub_map)

    conditional = Formula('->', unquantified_formula.first, new_witness)
    step3 = prover.add_tautological_implication(conditional, {step2, step1})
    step4 = prover.add_ug(Formula('A', variable, conditional), step3)
    step5 = prover.add_assumption(existential)

    inst_map = {'R': existential.statement.substitute({'x': Term('_')}),
                'Q': new_witness,
                'x': 'x'}
    step6 = prover.add_instantiated_assumption(Prover.ES.instantiate(inst_map), Prover.ES, inst_map)
    step7 = prover.add_tautological_implication(new_witness, {step4, step5, step6})
    step8 = prover.add_tautological_implication('~(~~P()->~~~~P())', {step1, step7})

    return prover.qed()
    # Task 12.7b

def existential_closure_step(sentences: AbstractSet[Formula]) -> Set[Formula]:
    """Augments the given sentences with an existential witness that uses a new
    constant name, for each existentially quantified sentence from these
    sentences for which an existential witness is missing.

    Parameters:
        sentences: prenex-normal-form sentences to augment with any missing
            existential witnesses.

    Returns:
        A set of all of the given sentences, and in addition for every
        existentially quantified sentence from the given sentences, a formula
        obtained from the statement quantified by that sentence by replacing all
        occurrences of the quantification variable name with a new constant name
        obtained by calling `next`(`~logic_utils.fresh_constant_name_generator`)`.
    """
    for sentence in sentences:
        assert is_in_prenex_normal_form(sentence) and \
               len(sentence.free_variables()) == 0
    constants = [Term(constant) for constant in get_constants(sentences)]
    new_sentences = set(sentences)
    for sentence in sentences:
        if sentence.root == 'E':
            if any(sentence.statement.substitute({sentence.variable: constant}) in sentences for constant in constants):
                continue
            else:
                new_sentences.add(sentence.statement.substitute({sentence.variable: Term(next(fresh_constant_name_generator))}))
    return new_sentences
    # Task 12.8


# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/syntax.py

"""Syntactic handling of predicate-logic expressions."""

from __future__ import annotations
from functools import lru_cache, reduce
from typing import AbstractSet, Mapping, Optional, Sequence, Set, Tuple, Union
from mltp.logic_utils import fresh_variable_name_generator, frozen, memoized_parameterless_method
from mltp.propositions.syntax import Formula as PropositionalFormula, is_variable as is_propositional_variable
# from predicates.operators import *

class ForbiddenVariableError(Exception):
    """Raised by `Term.substitute` and `Formula.substitute` when a substituted
    term contains a variable name that is forbidden in that context.

    Attributes:
        variable_name (`str`): the variable name that was forbidden in the
            context in which a term containing it was to be substituted.
    """
    variable_name: str

    def __init__(self, variable_name: str):
        """Initializes a `ForbiddenVariableError` from the offending variable
        name.

        Parameters:
            variable_name: variable name that is forbidden in the context in
                which a term containing it is to be substituted.
        """
        assert is_variable(variable_name)
        self.variable_name = variable_name

@lru_cache(maxsize=100) # Cache the return value of is_constant
def is_constant(string: str) -> bool:
    """Checks if the given string is a constant name.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a constant name, ``False`` otherwise.
    """
    return  (((string[0] >= '0' and string[0] <= '9') or \
              (string[0] >= 'a' and string[0] <= 'e')) and \
               string.isalnum()) or string == '_'

def make_constant(string:str, universe:set) -> str:
    """converts a string to a valid constant name.

    Parameters:
        string: string to convert.

    Returns:
        a string for which is_constant(string) will evaluate to True.
    """
    if not (string[0] >= '0' and string[0] <= '9') or \
              (string[0] >= 'a' and string[0] <= 'e') or \
               string[0] == '_':
        string = '0' + string
    if not string.isalnum():
        string = ''.join([i if i.isalnum() else '' for i in string])
    if string in universe:
        i=1
        new_string = string
        while new_string in universe:
            new_string = string+str(i)
            i+=1
        string = new_string
    return string
    # Personal task

@lru_cache(maxsize=100) # Cache the return value of is_variable
def is_variable(string: str) -> bool:
    """Checks if the given string is a variable name.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a variable name, ``False`` otherwise.
    """
    return string[0] >= 'u' and string[0] <= 'z' and string.isalnum()

@lru_cache(maxsize=100) # Cache the return value of is_function
def is_function(string: str) -> bool:
    """Checks if the given string is a function name.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a function name, ``False`` otherwise.
    """
    return string[0] >= 'f' and string[0] <= 't' and string.isalnum()

def separate_relation_name(string: str) -> Tuple:
    assert string and is_relation(string[0]), print(f'{string} does not start with a relation name')
    counter = 0
    while string[counter].isalnum():
        counter += 1
    return string[:counter], string[counter:]

def separate_bracketed_formula(string: str) -> Tuple:
    if (open_bracket := string[0]) not in {'[','('}:
        print(f"{string} doesn't start with a bracket")
        return string
    if open_bracket == '[':
        close_bracket =']'
    else:
        close_bracket = ')'
    counter = 0
    for i,j in enumerate(string):
        if j == open_bracket:
            counter +=1
        elif j == close_bracket:
            counter -=1
        if counter == 0:
            return string[:i+1], string[i+1:]

def separate_operator(string: str) -> list:
    if string[0] in {'&', '|', '+'}:
        return string[0], string[1:]
    elif string[0] == '-':
        assert string[1] in {'>','&','|'}
        return string[0:2], string[2:]
    else:
        assert string[0:3] == '<->'
        return string[0:3], string[3:]

def separate_term(string: str) -> Tuple:
    assert string
    sample = string[0]
    if sample == '_':
        return ('_', string[1:])
    counter = 0
    while counter < len(string) and string[counter].isalnum():
        counter += 1
    if is_constant(sample) or is_variable(sample):
        return string[:counter], string[counter:]

    elif is_function(sample):
        arguments, suffix = separate_bracketed_formula(string[counter:])
        return string[:counter]+arguments, suffix

def separate_function_or_relation_name(string: str) -> Tuple:
    assert string
    counter = 0
    while string[counter].isalnum():
        counter += 1
    return string[:counter], *separate_bracketed_formula(string[counter:])

@frozen
class Term:
    """An immutable predicate-logic term in tree representation, composed from
    variable names and constant names, and function names applied to them.

    Attributes:
        root (`str`): the constant name, variable name, or function name at the
            root of the term tree.
        arguments (`~typing.Optional`\\[`~typing.Tuple`\\[`Term`, ...]]): the
            arguments of the root, if the root is a function name.
    """
    root: str
    arguments: Optional[Tuple[Term, ...]]

    def __init__(self, root: str, arguments: Optional[Sequence[Term]] = None):
        """Initializes a `Term` from its root and root arguments.

        Parameters:
            root: the root for the formula tree.
            arguments: the arguments for the root, if the root is a function
                name.
        """
        if is_constant(root) or is_variable(root):
            assert arguments is None
            self.root = root
        else:
            assert is_function(root)
            assert arguments is not None and len(arguments) > 0
            self.root = root
            self.arguments = tuple(arguments)

    @memoized_parameterless_method
    def __repr__(self) -> str:
        """Computes the string representation of the current term.

        Returns:
            The standard string representation of the current term.
        """
        root = self.root
        if is_variable(root) or is_constant(root):
            return root
        else:
            assert type(self) == Term or type(self) == Formula
            return f'{root}({",".join([argument.__repr__() for argument in self.arguments])})'
        # Task 7.1

    def __eq__(self, other: object) -> bool:
        """Compares the current term with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is a `Term` object that equals the
            current term, ``False`` otherwise.
        """
        return isinstance(other, Term) and str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        """Compares the current term with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is not a `Term` object or does not
            equal the current term, ``False`` otherwise.
        """
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    @staticmethod
    def _parse_prefix(string: str) -> Tuple[Term, str]:
        """Parses a prefix of the given string into a term.

        Parameters:
            string: string to parse, which has a prefix that is a valid
                representation of a term.

        Returns:
            A pair of the parsed term and the unparsed suffix of the string. If
            the given string has as a prefix a constant name (e.g., ``'c12'``)
            or a variable name (e.g., ``'x12'``), then the parsed prefix will be
            that entire name (and not just a part of it, such as ``'x1'``).
        """
        if string:
            sample = string[0]
            if is_constant(sample) or is_variable(sample):
                prefix, suffix = separate_term(string)
                return Term(prefix), suffix
            else:
                assert is_function(sample)
                function_name, args, suffix = separate_function_or_relation_name(string)
                arg_suffix = args[1:-1]
                arg_list = []

                while arg_suffix:
                    arg_prefix, arg_suffix = separate_term(arg_suffix)
                    arg_list.append(arg_prefix)
                    if arg_suffix and arg_suffix[0] == ',':
                        arg_suffix = arg_suffix[1:]
                return Term(function_name, [Term._parse_prefix(arg)[0] for arg in arg_list]), suffix
        # Task 7.3a

    @staticmethod
    def parse(string: str) -> Term:
        """Parses the given valid string representation into a term.

        Parameters:
            string: string to parse.

        Returns:
            A term whose standard string representation is the given string.
        """
        term = Term._parse_prefix(string)[0]
        assert string == str(term), print(f'Ensure that the string {string} is a well-formed term')
        return term
        # Task 7.3b

    def constants(self) -> Set[str]:
        """Finds all constant names in the current term.

        Returns:
            A set of all constant names used in the current term.
        """
        root = self.root
        if is_constant(root):
            return {root}

        elif is_variable(root):
            return set()

        elif is_function(root):
            constants = set()
            for argument in self.arguments:
                constants = constants | Term.constants(argument)
            return constants
        # Task 7.5a

    def variables(self) -> Set[str]:
        """Finds all variable names in the current term.

        Returns:
            A set of all variable names used in the current term.
        """
        root = self.root
        if is_variable(root):
            return {root}

        elif is_constant(root):
            return set()

        elif is_function(root):
            variables = set()
            for argument in self.arguments:
                variables = variables | Term.variables(argument)
            return variables
        # Task 7.5b

    def functions(self) -> Set[Tuple[str, int]]:
        """Finds all function names in the current term, along with their
        arities.

        Returns:
            A set of pairs of function name and arity (number of arguments) for
            all function names used in the current term.
        """
        root = self.root
        functions = set()
        if is_function(root):
            functions = functions | {(root, len(self.arguments))}
            for argument in self.arguments:
                functions = functions | Term.functions(argument)
        return functions
        # Task 7.5c

    def __len__(self) -> int:
        return len(self.variables()) + len(self.constants()) + len(self.functions())
        # personal task

    def substitute(self, substitution_map: Mapping[str, Term],
                   forbidden_variables: AbstractSet[str] = frozenset()) -> Term:
        """Substitutes in the current term, each constant name `construct` or
        variable name `construct` that is a key in `substitution_map` with the
        term `substitution_map[`construct`]`.

        Parameters:
            substitution_map: mapping defining the substitutions to be
                performed.
            forbidden_variables: variable names not allowed in substitution
                terms.

        Returns:
            The term resulting from performing all substitutions. Only
            constant name and variable name occurrences originating in the
            current term are substituted (i.e., those originating in one of the
            specified substitutions are not subjected to additional
            substitutions).

        Raises:
            ForbiddenVariableError: If a term that is used in the requested
                substitution contains a variable name from
                `forbidden_variables`.

        Examples:
            >>> Term.parse('f(x,c)').substitute(
            ...     {'c': Term.parse('plus(d,x)'), 'x': Term.parse('c')}, {'y'})
            f(c,plus(d,x))

            >>> Term.parse('f(x,c)').substitute(
            ...     {'c': Term.parse('plus(d,y)')}, {'y'})
            Traceback (most recent call last):
              ...
            predicates.syntax.ForbiddenVariableError: y
        """
        for construct in substitution_map:
            assert is_constant(construct) or is_variable(construct)
        for variable in forbidden_variables:
            assert is_variable(variable)

        if not substitution_map:
            return self

        if sinners:= set.union(*[value.variables() for value in substitution_map.values()]) & forbidden_variables:
            raise ForbiddenVariableError(sinners.pop())

        string = str(self)
        if string in substitution_map:
            term = substitution_map[string]
        elif is_variable(string) or is_constant(string):
            term = self
        else:
            term = Term(self.root, [argument.substitute(substitution_map, forbidden_variables) for argument in self.arguments])
        return term
        # Task 9.1

@lru_cache(maxsize=100) # Cache the return value of is_equality
def is_equality(string: str) -> bool:
    """Checks if the given string is the equality relation.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is the equality relation, ``False``
        otherwise.
    """
    return string == '='

@lru_cache(maxsize=100) # Cache the return value of is_relation
def is_relation(string: str) -> bool:
    """Checks if the given string is a relation name.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a relation name, ``False`` otherwise.
    """
    return string[0] >= 'F' and string[0] <= 'T' and string.isalnum()

@lru_cache(maxsize=100) # Cache the return value of is_unary
def is_unary(string: str) -> bool:
    """Checks if the given string is a unary operator.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a unary operator, ``False`` otherwise.
    """
    return string == '~'

@lru_cache(maxsize=100) # Cache the return value of is_binary
def is_binary(string: str) -> bool:
    """Checks if the given string is a binary operator.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a binary operator, ``False`` otherwise.
    """
    return string in {'&', '|', '->', '<->', '+', '-|', '-&'}

@lru_cache(maxsize=100) # Cache the return value of is_quantifier
def is_quantifier(string: str) -> bool:
    """Checks if the given string is a quantifier.

    Parameters:
        string: string to check.

    Returns:
        ``True`` if the given string is a quantifier, ``False`` otherwise.
    """
    return string == 'A' or string == 'E'

@frozen
class Formula:
    """An immutable predicate-logic formula in tree representation, composed
    from relation names applied to predicate-logic terms, and operators and
    quantifications applied to them.

    Attributes:
        root (`str`): the relation name, equality relation, operator, or
            quantifier at the root of the formula tree.
        arguments (`~typing.Optional`\\[`~typing.Tuple`\\[`Term`, ...]]): the
            arguments of the root, if the root is a relation name or the
            equality relation.
        first (`~typing.Optional`\\[`Formula`]): the first operand of the root,
            if the root is a unary or binary operator.
        second (`~typing.Optional`\\[`Formula`]): the second operand of the
            root, if the root is a binary operator.
        variable (`~typing.Optional`\\[`str`]): the variable name quantified by
            the root, if the root is a quantification.
        statement (`~typing.Optional`\\[`Formula`]): the statement quantified by
            the root, if the root is a quantification.
    """
    root: str
    arguments: Optional[Tuple[Term, ...]]
    first: Optional[Formula]
    second: Optional[Formula]
    variable: Optional[str]
    statement: Optional[Formula]

    def __init__(self, root: str,
                 arguments_or_first_or_variable: Union[Sequence[Term],
                                                       Formula, str],
                 second_or_statement: Optional[Formula] = None):
        """Initializes a `Formula` from its root and root arguments, root
        operands, or root quantified variable name and statement.

        Parameters:
            root: the root for the formula tree.
            arguments_or_first_or_variable: the arguments for the root, if the
                root is a relation name or the equality relation; the first
                operand for the root, if the root is a unary or binary operator;
                the variable name to be quantified by the root, if the root is a
                quantification.
            second_or_statement: the second operand for the root, if the root is
                a binary operator; the statement to be quantified by the root,
                if the root is a quantification.
        """
        if is_equality(root) or is_relation(root):
            # Populate self.root and self.arguments
            assert isinstance(arguments_or_first_or_variable, Sequence) and not isinstance(arguments_or_first_or_variable, str)
            if is_equality(root):
                assert len(arguments_or_first_or_variable) == 2
            assert second_or_statement is None
            self.root, self.arguments = root, tuple(argument if isinstance(argument, Term) else Term(argument) for argument in arguments_or_first_or_variable)

        elif is_unary(root):
            # Populate self.first
            assert isinstance(arguments_or_first_or_variable, Formula), print(f'{arguments_or_first_or_variable} is {type(arguments_or_first_or_variable)}, should be Formula')
            assert second_or_statement is None
            self.root, self.first = root, arguments_or_first_or_variable

        elif is_binary(root):
            # Populate self.first and self.second
            assert isinstance(arguments_or_first_or_variable, Formula), print(f'{arguments_or_first_or_variable} is {type(arguments_or_first_or_variable)}')
            assert second_or_statement is not None
            self.root, self.first, self.second = \
                root, arguments_or_first_or_variable, second_or_statement

        else:
            assert is_quantifier(root)
            # Populate self.variable and self.statement
            assert isinstance(arguments_or_first_or_variable, str) and \
                   is_variable(arguments_or_first_or_variable)
            assert second_or_statement is not None
            self.root, self.variable, self.statement = root, arguments_or_first_or_variable, second_or_statement

    @memoized_parameterless_method
    def __repr__(self) -> str:
        """Computes the string representation of the current formula.

        Returns:
            The standard string representation of the current formula.
        """
        root = self.root
        if is_equality(root):
            return f'{self.arguments[0].__repr__()}{root}{self.arguments[1].__repr__()}'

        elif is_relation(root):
            return f'{root}({",".join([argument.__repr__() for argument in self.arguments])})'

        elif is_unary(root):
            return f'{root}{self.first.__repr__()}'

        elif is_binary(root):
            return f'({self.first.__repr__()}{root}{self.second.__repr__()})'

        else:
            assert is_quantifier(root)
            return f'{root}{self.variable}[{self.statement.__repr__()}]'

        # Task 7.2

    def __eq__(self, other: object) -> bool:
        """Compares the current formula with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is a `Formula` object that equals the
            current formula, ``False`` otherwise.
        """
        return isinstance(other, Formula) and str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        """Compares the current formula with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is not a `Formula` object or does not
            equal the current formula, ``False`` otherwise.
        """
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    @staticmethod
    def _parse_prefix(string: str) -> Tuple[Formula, str]:
        """Parses a prefix of the given string into a formula.

        Parameters:
            string: string to parse, which has a prefix that is a valid
                representation of a formula.

        Returns:
            A pair of the parsed formula and the unparsed suffix of the string.
            If the given string has as a prefix a term followed by an equality
            followed by a constant name (e.g., ``'f(y)=c12'``) or by a variable
            name (e.g., ``'f(y)=x12'``), then the parsed prefix will include
            that entire name (and not just a part of it, such as ``'f(y)=x1'``).
        """
        sample = string[0]
        if is_constant(sample) or is_variable(sample) or is_function(sample):
            first, suffix = separate_term(string)
            assert suffix[0] == '='
            second, suffix = separate_term(suffix[1:])
            formula = Formula('=', [Term.parse(first), Term.parse(second)])

        elif is_relation(sample):
            relation_name, args, suffix = separate_function_or_relation_name(string)
            arg_suffix = args[1:-1]
            arg_list = []

            while arg_suffix:
                arg_prefix, arg_suffix = separate_term(arg_suffix)
                arg_list.append(arg_prefix)
                if arg_suffix and arg_suffix[0] == ',':
                    arg_suffix = arg_suffix[1:]
            formula = Formula(relation_name, [Term._parse_prefix(arg)[0] for arg in arg_list])

        elif is_unary(sample):
            prefix, suffix = Formula._parse_prefix(string[1:])
            formula = Formula('~', prefix)

        elif is_quantifier(sample):
            quantifier = sample
            variable, suffix = separate_term(string[1:])
            statement, suffix = separate_bracketed_formula(suffix)
            formula = Formula(quantifier, variable, Formula._parse_prefix(statement[1:-1])[0])

        else:
            assert sample == '('
            binary_formula, suffix = separate_bracketed_formula(string)
            first, remaining = Formula._parse_prefix(binary_formula[1:])
            operator, second = separate_operator(remaining)
            formula = Formula(operator, first, Formula._parse_prefix(second)[0])
        return formula, suffix
        # Task 7.4a

    @staticmethod
    def parse(string: str) -> Formula:
        """Parses the given valid string representation into a formula.

        Parameters:
            string: string to parse.

        Returns:
            A formula whose standard string representation is the given string.
        """
        formula = Formula._parse_prefix(string)[0]
        assert string == str(formula), print(f"Ensure that string {string} is a well-formed formula. It doesn't match {str(formula)}")
        return formula
        # Task 7.4b

    def constants(self) -> Set[str]:
        """Finds all constant names in the current formula.
        Returns:
            A set of all constant names used in the current formula.
        """
        root = self.root
        if is_quantifier(root):
            return Formula.constants(self.statement)

        elif is_unary(root):
            return Formula.constants(self.first)

        elif is_binary(root):
            return Formula.constants(self.first) | Formula.constants(self.second)

        elif is_equality(root) or is_relation(root):
            constants = set()
            for argument in self.arguments:
                constants = constants | Term.constants(argument)
            return constants
        # Task 7.6a

    def variables(self) -> Set[str]:
        """Finds all variable names in the current formula.

        Returns:
            A set of all variable names used in the current formula.
        """
        root = self.root
        if is_quantifier(root):
            return Formula.variables(self.statement) | {self.variable}

        elif is_unary(root):
            return Formula.variables(self.first)

        elif is_binary(root):
            return Formula.variables(self.first) | Formula.variables(self.second)

        elif is_equality(root) or is_relation(root):
            variables = set()
            for argument in self.arguments:
                variables = variables | Term.variables(argument)
            return variables
        # Task 7.6b

    def free_variables(self) -> Set[str]:
        """Finds all variable names that are free in the current formula.

        Returns:
            A set of every variable name that is used in the current formula not
            only within a scope of a quantification on that variable name.
        """
        root = self.root
        if is_quantifier(root):
            free_variables = Formula.free_variables(self.statement) - {self.variable}
            return free_variables

        elif is_unary(root):
            return Formula.free_variables(self.first)

        elif is_binary(root):
            return Formula.free_variables(self.first) | Formula.free_variables(self.second)

        elif is_equality(root) or is_relation(root):
            free_variables = set()
            for argument in self.arguments:
                free_variables = free_variables | Term.variables(argument)
            return free_variables
        # Task 7.6c

    def bound_variables(self) -> Set[str]:
        """Finds all variable names that are quantified in the current formula.

        Returns:
            A set of every variable name that is used in the current formula
            within a scope of a quantification on that variable name.
        """
        root = self.root
        if is_quantifier(root):
            return {self.variable} | self.statement.bound_variables()

        elif is_unary(root):
            return Formula.bound_variables(self.first)

        elif is_binary(root):
            return Formula.bound_variables(self.first) | Formula.bound_variables(self.second)

        elif is_equality(root) or is_relation(root):
            return set()

        # Personal task

    def functions(self) -> Set[Tuple[str, int]]:
        """Finds all function names in the current formula, along with their
        arities.

        Returns:
            A set of pairs of function name and arity (number of arguments) for
            all function names used in the current formula.
        """
        root = self.root
        if is_quantifier(root):
            return Formula.functions(self.statement)

        elif is_unary(root):
            return Formula.functions(self.first)

        elif is_binary(root):
            return Formula.functions(self.first) | Formula.functions(self.second)

        elif is_equality(root) or is_relation(root):
            functions = set()
            for argument in self.arguments:
                functions = functions | Term.functions(argument)
            return functions
        # Task 7.6d

    def relations(self) -> Set[Tuple[str, int]]:
        """Finds all relation names in the current formula, along with their
        arities.

        Returns:
            A set of pairs of relation name and arity (number of arguments) for
            all relation names used in the current formula.
        """
        root = self.root
        if is_relation(root):
            return {(root, len(self.arguments))}

        elif is_quantifier(root):
            return Formula.relations(self.statement)

        elif is_unary(root):
            return Formula.relations(self.first)

        elif is_binary(root):
            return Formula.relations(self.first) | Formula.relations(self.second)

        return set()
        # Task 7.6e

    def operators(self) -> Set[Tuple[str, int]]:
        """Finds all relation names in the current formula, along with their
        arities.

        Returns:
            A set of pairs of relation name and arity (number of arguments) for
            all relation names used in the current formula.
        """
        root = self.root
        if is_quantifier(root):
            return Formula.operators(self.statement)

        elif is_unary(root):
            return {root} | Formula.operators(self.first)

        elif is_binary(root):
            return {root} | Formula.operators(self.first) | Formula.operators(self.second)

        return set()
        # Personal Task

    def quantifiers(self) -> Set[Tuple[str, str]]:
        """Finds all quantifiers in the current formula.

        Returns:
            A set of tuples of every quantifier with its associated variable name that is used in the current formula.
        """
        root = self.root
        if is_quantifier(root):
            return {(root, self.variable)} | self.statement.quantifiers()

        elif is_unary(root):
            return Formula.quantifiers(self.first)

        elif is_binary(root):
            return Formula.quantifiers(self.first) | Formula.quantifiers(self.second)

        elif is_equality(root) or is_relation(root):
            return set()
        # personal task

    def __len__(self) -> int:
        return len(self.constants()) + len(self.variables()) + len(self.quantifiers()) + len(self.functions()) + len(self.relations())
        # personal task

    def substitute(self, substitution_map: Mapping[str, Term],
                   forbidden_variables: AbstractSet[str] = frozenset()) -> \
            Formula:
        """Substitutes in the current formula, each constant name `construct` or
        free occurrence of variable name `construct` that is a key in
        `substitution_map` with the term `substitution_map[construct]`.

        Parameters:
            substitution_map: mapping defining the substitutions to be
                performed.
            forbidden_variables: variable names not allowed in substitution
                terms.

        Returns:
            The formula resulting from performing all substitutions. Only
            constant name and variable name occurrences originating in the
            current formula are substituted (i.e., those originating in one of
            the specified substitutions are not subjected to additional
            substitutions).

        Raises:
            ForbiddenVariableError: If a term that is used in the requested
                substitution contains a variable name from `forbidden_variables`
                or a variable name occurrence that becomes bound when that term
                is substituted into the current formula.

        Examples:
            >>> Formula.parse('Ay[x=c]').substitute(
            ...     {'c': Term.parse('plus(d,x)'), 'x': Term.parse('c')}, {'z'})
            Ay[c=plus(d,x)]

            >>> Formula.parse('Ay[x=c]').substitute(
            ...     {'c': Term.parse('plus(d,z)')}, {'z'})
            Traceback (most recent call last):
              ...
            predicates.syntax.ForbiddenVariableError: z

            >>> Formula.parse('Ay[x=c]').substitute(
            ...     {'c': Term.parse('plus(d,y)')})
            Traceback (most recent call last):
              ...
            predicates.syntax.ForbiddenVariableError: y
        """
        for construct in substitution_map:
            assert is_constant(construct) or is_variable(construct)
        for variable in forbidden_variables:
            assert is_variable(variable)

        if not set(substitution_map.keys()) & (self.variables() | self.constants()):
            return self

        if sinners:= set.union(*[value.variables() for value in substitution_map.values()]) & forbidden_variables:
            raise ForbiddenVariableError(sinners.pop())

        if str(self) in substitution_map:
            formula = substitution_map[str(self)]
        else:
            root = self.root
            if is_unary(root):
               formula = Formula(root,
                        self.first.substitute(substitution_map, forbidden_variables))

            elif is_binary(root):
                formula = Formula(root,
                        self.first.substitute(substitution_map, forbidden_variables),
                        self.second.substitute(substitution_map, forbidden_variables))

            elif is_relation(root) or is_equality(root):
                formula = Formula(root,
                        [argument.substitute(substitution_map, forbidden_variables) for argument in self.arguments])

            else:
                forbidden_variables = set(forbidden_variables) | {self.variable}
                new_substitution_map = {key:substitution_map[key] for key in substitution_map if not key == self.variable}
                if new_substitution_map:
                    formula = Formula(root,
                            self.variable,
                            self.statement.substitute(new_substitution_map, forbidden_variables))
                else:
                    formula = self
        return formula
        # Task 9.2

    def propositional_skeleton(self, substitution_map=None) -> Tuple[PropositionalFormula,
                                              Mapping[str, Formula]]:
        """Computes a propositional skeleton of the current formula.

        Returns:
            A pair. The first element of the pair is a propositional formula
            obtained from the current formula by substituting every (outermost)
            subformula that has a relation name, equality, or quantifier at its
            root with a propositional variable name, consistently such that
            multiple identical such (outermost) subformulas are substituted with
            the same propositional variable name. The propositional variable
            names used for substitution are obtained, from left to right
            (considering their first occurrence), by calling
            `next(~logic_utils.fresh_variable_name_generator)`.
            The second element of the pair is a mapping from each propositional
            variable name to the subformula for which it was substituted.

        Examples:
            >>> formula = Formula.parse('((Ax[x=7]&x=7)|(~Q(y)->x=7))')
            >>> formula.propositional_skeleton()
            (((z1&z2)|(~z3->z2)), {'z1': Ax[x=7], 'z2': x=7, 'z3': Q(y)})
            >>> formula.propositional_skeleton()
            (((z4&z5)|(~z6->z5)), {'z4': Ax[x=7], 'z5': x=7, 'z6': Q(y)})
        """
        if not substitution_map:
            substitution_map = dict()
        root = self.root
        if is_unary(root):
            first, substitution_map = self.first.propositional_skeleton(substitution_map)
            formula = PropositionalFormula(root, first)

        elif is_binary(root):
            first, substitution_map = self.first.propositional_skeleton(substitution_map)
            if self.second in substitution_map.values():
                for _tuple in substitution_map.items():
                    if self.second in _tuple:
                        second = PropositionalFormula.parse(_tuple[0])
                        break
            else:
                second, substitution_map2 = self.second.propositional_skeleton(substitution_map)
                substitution_map.update(substitution_map2)
            formula = PropositionalFormula(root, first, second)

        elif is_relation(root) or is_equality(root) or is_quantifier(root):
            if self in substitution_map.values():
                for _tuple in substitution_map.items():
                    if self in _tuple:
                        formula = PropositionalFormula(_tuple[0])
                        break
            else:
                variable = next(fresh_variable_name_generator)
                substitution_map[variable] = self
                formula = PropositionalFormula.parse(variable)
        return formula, substitution_map
        # Task 9.8

    @staticmethod
    def from_propositional_skeleton(skeleton: PropositionalFormula,
                                    substitution_map: Mapping[str, Formula]) \
            -> Formula:
        """Computes a predicate-logic formula from a propositional skeleton and
        a substitution map.

        Arguments:
            skeleton: propositional skeleton for the formula to compute,
                containing no constants or operators beyond '~', '->',
                '|', '&', '+', '<->', '-|', and '-&'.
            substitution_map: mapping from each propositional variable name of
                the given propositional skeleton to a predicate-logic formula.

        Returns:
            A predicate-logic formula obtained from the given propositional
            skeleton by substituting each propositional variable name with the
            formula mapped to it by the given map.

        Examples:
            >>> Formula.from_propositional_skeleton(
            ...     PropositionalFormula.parse('((z1&z2)|(~z3->z2))'),
            ...     {'z1': Formula.parse('Ax[x=7]'), 'z2': Formula.parse('x=7'),
            ...      'z3': Formula.parse('Q(y)')})
            ((Ax[x=7]&x=7)|(~Q(y)->x=7))

            >>> Formula.from_propositional_skeleton(
            ...     PropositionalFormula.parse('((z9&z2)|(~z3->z2))'),
            ...     {'z2': Formula.parse('x=7'), 'z3': Formula.parse('Q(y)'),
            ...      'z9': Formula.parse('Ax[x=7]')})
            ((Ax[x=7]&x=7)|(~Q(y)->x=7))
        """
        for operator in skeleton.operators():
            assert is_unary(operator) or is_binary(operator)
        for variable in skeleton.variables():
            assert variable in substitution_map

        root = skeleton.root
        if is_propositional_variable(root) and root in substitution_map:
            return substitution_map[root]
        else:
            if is_unary(root):
               return Formula(root, Formula.from_propositional_skeleton(skeleton.first, substitution_map))
            elif is_binary(root):
                return Formula(root,
                    Formula.from_propositional_skeleton(skeleton.first, substitution_map),
                    Formula.from_propositional_skeleton(skeleton.second, substitution_map))
        # Task 9.10



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/axiomatic_systems.py

"""Axiomatic inference rules of propositional logic."""
from .syntax import *
from .proofs import *

# Axiomatic inference rules that only contain implies

#: Modus ponens / implication elimination
MP = InferenceRule([Formula.parse('p'), Formula.parse('(p->q)')],
                   Formula.parse('q'))
#: Self implication
I0 = InferenceRule([], Formula.parse('(p->p)'))
#: Implication introduction (right)
I1 = InferenceRule([], Formula.parse('(q->(p->q))'))
#: Self-distribution of implication
D = InferenceRule([], Formula.parse('((p->(q->r))->((p->q)->(p->r)))'))

# Axiomatic inference rules for not (and implies)

#: Implication introduction (left)
I2 = InferenceRule([], Formula.parse('(~p->(p->q))'))
#: Converse contraposition
N = InferenceRule([], Formula.parse('((~q->~p)->(p->q))'))
#: Negative-implication introduction
NI = InferenceRule([], Formula.parse('(p->(~q->~(p->q)))'))
#: Double-negation introduction
NN = InferenceRule([], Formula.parse('(p->~~p)'))
#: Resolution
R = InferenceRule([], Formula.parse('((q->p)->((~q->p)->p))'))

#: Large axiomatic system for implication and negation, consisting of `MP`,
#: `I0`, `I1`, `D`, `I2`, `N`, `NI`, `NN`, and `R`.
AXIOMATIC_SYSTEM = frozenset({MP, I0, I1, D, I2, N, NI, NN, R})
#: Hilbert axiomatic system for implication and negation, consisting of `MP`,
#: `I1`, `D`, and `N`.
HILBERT_AXIOMATIC_SYSTEM = frozenset({MP, I1, D, N})

# Axiomatic inference rules for conjunction (and implication and negation)

#: Conjunction introduction
A = InferenceRule([], Formula.parse('(p->(q->(p&q)))'))
#: Negative-conjunction introduction (right)
NA1 = InferenceRule([], Formula.parse('(~q->~(p&q))'))
#: Negative-conjunction introduction (left)
NA2 = InferenceRule([], Formula.parse('(~p->~(p&q))'))

# Axiomatic inference rules for disjunction (and implication and negation)

# Disjunction introduction (right)
O1 = InferenceRule([], Formula.parse('(q->(p|q))'))
# Disjunction introduction (left)
O2 = InferenceRule([], Formula.parse('(p->(p|q))'))
# Negative-disjunction introduction
NO = InferenceRule([], Formula.parse('(~p->(~q->~(p|q)))'))

# Axiomatic inference rules for constants (and implication and negation)

#: Truth introduction
T =  InferenceRule([], Formula.parse('T'))
#: Negative-falsity introduction
NF = InferenceRule([], Formula.parse('~F'))

#: Large axiomatic system for all operators, consisting of the rules in
#: `AXIOMATIC_SYSTEM`, as well as `A`, `NA1`, `NA2`, `O1`, `O2`, `NO`, `T`, and
#: `NF`.
AXIOMATIC_SYSTEM_FULL = \
    frozenset(AXIOMATIC_SYSTEM.union({A, NA1, NA2, O1, O2, NO, T, NF}))

# Alternative for N

#: Reductio ad absurdum
N_ALTERNATIVE = InferenceRule([], Formula.parse('((~q->~p)->((~q->p)->q))'))

#: Hilbert axiomatic system for implication and negation, with `N` replaced by
#: `N_ALTERNATIVE`.
HILBERT_AXIOMATIC_SYSTEM_ALTERNATIVE = frozenset({MP, I1, D, N_ALTERNATIVE})

# Alternatives for NA1, NA2, NO without negation

#: Conjunction elimination (right)
AE1 = InferenceRule([], Formula.parse('((p&q)->q)'))
#: Conjunction elimination (left)
AE2 = InferenceRule([], Formula.parse('((p&q)->p)'))
#: Disjunction elimination
OE = InferenceRule([], Formula.parse('((p->r)->((q->r)->((p|q)->r)))'))

#: Hilbert axiomatic system for all operators, consisting of the rules in
#: `HILBERT_AXIOMATIC_SYSTEM`, as well as `A`, `AE1`, `AE2`, `O1`, `O2`, `OE`,
#: `T`, and `NF`.
HILBERT_AXIOMATIC_SYSTEM_FULL = \
    frozenset(HILBERT_AXIOMATIC_SYSTEM.union({A, AE1, AE2, O1, O2, OE, T, NF}))
